/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.framework.engine.source;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import java.security.Principal;
import java.sql.Timestamp;

import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.SyncSource;

/*
 * This source represents a <i>SyncSource</i> where all items are kept in memory.
 * A <i>MemorySyncSource</i> can be created passing in the constructor the name
 * of the source and the items or creating an anonymous source with the default
 * constructor and then initialize the source with <i>initialize(...)</i>.
 * <p>
 * <code>existingSyncItems</code> represents the items on the device.
 * <p>
 * <code>newSyncItems</code>, <code>updatedSyncItems</code> and 
 * <code>deletedSyncItems</code> are insted of the modifications reported with 
 * the last message.
 *
 * @author Stefano Fornari @ Funambol
 *
 * @version $Id: MemorySyncSource.java,v 1.11.4.3 2005/01/26 16:55:04 stefano_fornari Exp $
 */
public class MemorySyncSource 
extends AbstractSyncSource
implements SyncSource, java.io.Serializable {
    
    // --------------------------------------------------------------- Constants
    
    public static final String NAME = "MemorySyncSource";
    
    // ----------------------------------------------------- Private data member
    
    // the content of the SyncML client
    private List existingSyncItems;
    private List newSyncItems     ;
    private List updatedSyncItems ;
    private List deletedSyncItems ;
    
    /**
     * LO SyncItem - not null if any LO is being retrieved
     */
    private SyncItem lo;
    
    // ------------------------------------------------------------ Constructors
    
    public MemorySyncSource() {
        lo = null;
    }
    
    public MemorySyncSource(String name, String type, String sourceURI) {
        super(name, type, sourceURI);
        lo = null;
        initialize(new ArrayList(), new ArrayList(), new ArrayList());
        setExistingItems(new ArrayList());
    }
    
    public MemorySyncSource(String name           , 
                            List unchangedSyncItems,
                            List deletedSyncItems ,
                            List newSyncItems     ) {
      //
      // The items are added so no need to clone the lists
      //
      super(name);
      lo = null;
      initialize(deletedSyncItems, newSyncItems, updatedSyncItems);
      setExistingItems(new ArrayList());
    }
    
    // ---------------------------------------------------------- Public Methods
    
    /**
     * Sets the name of the source.
     *
     * @param name the new name
     */
    public void setName(String name) {
        this.name = name;
    }
    
    /*
    * @see SyncSource
    */

    public SyncItem setSyncItem(Principal principal, SyncItem syncItem) 
    throws SyncSourceException {
        newSyncItems.remove (syncItem);
        deletedSyncItems.add(syncItem);
        
        updatedSyncItems.add(syncItem);
        
        SyncItemKey mappedKey = syncItem.getMappedKey();
        String mappedKeyValue = (mappedKey == null) ? null
                                                    : mappedKey.getKeyAsString();
        
        SyncItem newSyncItem = new SyncItemImpl(
                                   this, 
                                   syncItem.getKey().getKeyAsString(), 
                                   mappedKeyValue,
                                   SyncItemState.NEW
                               );

        newSyncItem.setProperties(syncItem.getProperties());

        return newSyncItem;
    }
    
    /*
    * @see SyncSource
    */

    public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems) 
    throws SyncSourceException {
        SyncItem[] ret = new SyncItem[syncItems.length];
        for (int i=0; (i<syncItems.length); ++i) {
            ret[i] = setSyncItem(principal,  syncItems[i]);
        }  // next i
        
        return ret;
    }

    /*
     * @see SyncSource
     * 
     * This implementation cycles through all SyncItems looking for the 
     * specified key.
     */
    public SyncItem getSyncItemFromId(Principal principal, SyncItemKey syncItemKey) 
    throws SyncSourceException {
        SyncItem[] all = this.getAllSyncItems(principal);

        for (int i=0; ((all != null) && (i<all.length)); ++i) {
            if (syncItemKey.equals(all[i].getKey())) {
                return all[i];
            }
        }
            
        return null;
    }
    
    /*
     * @see SyncSource
     * 
     */
    public SyncItem[] getSyncItemsFromIds(Principal principal, SyncItemKey[] syncItemKeys) 
    throws SyncSourceException {
        ArrayList ret = new ArrayList();
        SyncItem  syncItem = null;
        
        for (int i=0; ((syncItemKeys != null) && (i<syncItemKeys.length)); ++i) {
            syncItem = getSyncItemFromId(principal, syncItemKeys[i]);
            
            if (syncItem != null) {
                ret.add(syncItem);
            }
        }
            
        return (SyncItem[])ret.toArray(new SyncItem[ret.size()]);
    }
    
    /*
    * @see SyncSource
    */
    public SyncItemKey[] getNewSyncItemKeys(Principal principal,
                                            Timestamp since    ) 
    throws SyncSourceException {
        return extractKeys(newSyncItems);
    }
    
    /*
    * @see SyncSource
    */
    public SyncItem[] getNewSyncItems(Principal principal,
                                      Timestamp since    ) 
    throws SyncSourceException {
        return (SyncItem[])newSyncItems.toArray(new SyncItem[0]);
    }

    /*
    * @see SyncSource
    */
    public SyncItemKey[] getDeletedSyncItemKeys(Principal principal,
                                                Timestamp since    ) 
    throws SyncSourceException {
        return extractKeys(deletedSyncItems);
    }
    
    /*
    * @see SyncSource
    */
    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Timestamp since    ) 
    throws SyncSourceException {
        return (SyncItem[])deletedSyncItems.toArray(new SyncItem[0]);
    }
    
    /**
     * @return an array of keys containing the <i>SyncItem</i>'s key of the updated
     *        items after the last synchronizazion.
     */
    public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,
                                                Timestamp since    ) 
    throws SyncSourceException {
        return extractKeys(updatedSyncItems);
    }
    
    /**
    * @see SyncSource
    */
    public SyncItem[] getUpdatedSyncItems(Principal principal,
                                          Timestamp since    )
    throws SyncSourceException {
        return (SyncItem[])updatedSyncItems.toArray(new SyncItem[0]);
    }

    /*
    * @see SyncSource
    */
    public void removeSyncItem(Principal principal, SyncItem syncItem) 
    throws SyncSourceException {
        newSyncItems.remove      (syncItem);
        updatedSyncItems.remove  (syncItem);
            
        deletedSyncItems.add(syncItem);
    }
    
    /*
    * @see SyncSource
    */
    public void removeSyncItems(Principal principal, SyncItem[] syncItems) 
    throws SyncSourceException {
        for (int i=0; i<syncItems.length; ++i) {
            removeSyncItem(principal, syncItems[i]);
        } // next i
    }
    
    /**
    * @see SyncSource
    */
    public SyncItem[] getAllSyncItems(Principal principal) 
    throws SyncSourceException {
        return (SyncItem[])existingSyncItems.toArray(new SyncItemImpl[0]);
    }
    
    /**
     * @see SyncSource
     */
    public boolean isModified(Principal principal, Timestamp since) {
        try {
            return (getDeletedSyncItems(principal, since).length > 0)
                || (getNewSyncItems(principal, since).length     > 0)
                || (getUpdatedSyncItems(principal, since).length > 0);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }
    
    /**
     * Initializes this MemorySyncSource with items.
     *
     * @param deletedItems the deleted items
     * @param newItems the created items
     * @param updatedItems the updated items
     */
    public void initialize(List deletedItems ,
                           List newItems     ,
                           List updatedItems ) {
       
       //
       // Large object handling
       // =====================
       // If the server is still retrieving a large object, lo will be set to 
       // the item being retrieved and the new piece of data must be merged to 
       // the previous content. The partial item must be then removed by the 
       // list that contains it. When the LO is fully retrieved, the complete 
       // item will be updated in the containing list and from now on it will 
       // like any other item. The only difference is that the property 
       // PROPERTY_SIZE will be set to the expected content size
       //
       if (deletedSyncItems == null) {
           deletedSyncItems = new ArrayList();
       } else {
           deletedSyncItems.clear();
       }
       
       if (newSyncItems == null) {
           newSyncItems = new ArrayList();
       } else {
           newSyncItems.clear();
       }
       
       if (updatedSyncItems == null) {
           updatedSyncItems = new ArrayList();
       } else {
           updatedSyncItems.clear();
       }
       
       deletedSyncItems.addAll(deletedItems)  ;
       newSyncItems.addAll(newItems)          ;
       updatedSyncItems.addAll(updatedItems ) ;
       
       handleLO();
       
       //
       // We have to make sure that the complete view is consistent so that
       // new, modified or deleted items are not in the unchanged list; that
       // deleted items are not in new, update and unchanged lists, and so on.
       //
       SyncItem syncItem = null;
       Iterator i = deletedSyncItems.iterator();
       while (i.hasNext()) {
           syncItem = (SyncItem)i.next();
           newSyncItems.remove(syncItem)      ;
           updatedSyncItems.remove(syncItem)  ;
       }
       
       i = updatedSyncItems.iterator();
       while (i.hasNext()) {
           syncItem = (SyncItem)i.next();           
           newSyncItems.remove(syncItem)      ;
       }
       
       i = newSyncItems.iterator();
       while (i.hasNext()) {
           syncItem = (SyncItem)i.next();
       }
    }
    
    
    /**
     * Sets the existing items
     *
     * @param items existing items
     *
     */
    public void setExistingItems(List items) {
        existingSyncItems = items;
    }
    
    
    /** 
     * Search the items for an item with exactly the same binary content.
     *
     * @return an item from a twin item. Each source implementation is free to
     *         interpret this as it likes (i.e.: comparing all fields).
     *
     * @param principal
     * @param syncItemTwin the twin item
     *
     * @throws SyncSourceException in case of errors (for instance if the
     *         underlying data store runs into problems)
     *
     */
    public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItemTwin) 
    throws SyncSourceException {
        SyncItem[] syncItems = 
            getSyncItemsFromTwins(principal, new SyncItem[] {syncItemTwin});
            
        if ((syncItems == null) || (syncItems.length == 0)) {
            return null; // not found
        }

        return syncItems[0];
    }
    
    /** 
     * Two items are considered twin if their binary content, converted to 
     * String are equal.
     *
     * @param principal 
     * @param syncItemTwins the twin items
     *
     * @return an array of items from corresponding twin items. 
     *
     * @throws SyncSourceException in case of errors (for instance if the
     *         underlying data store runs into problems)
     *
     */
    public SyncItem[] getSyncItemsFromTwins(Principal principal, SyncItem[] syncItemTwins) 
    throws SyncSourceException {
        ArrayList ret = new ArrayList();
        
        String[] contents = new String[syncItemTwins.length];
        
        for (int i=0; i<syncItemTwins.length; ++i) {
            contents[i] = new String((byte[])syncItemTwins[i].getProperty(SyncItem.PROPERTY_BINARY_CONTENT).getValue());
        }
        
        SyncItem[] all = getAllSyncItems(principal);
        
        String content = null;
        SyncProperty prop = null;
        for (int i=0; i<all.length; ++i) {
            prop = all[i].getProperty(SyncItem.PROPERTY_BINARY_CONTENT);
            if (prop != null) {
                content = new String((byte[])prop.getValue());
            }
            
            for (int j=0; ((content != null) && (j<contents.length)); ++j) {
                if (content.equals(contents[j])) {
                    ret.add(all[i]);
                    break;
                }
            }
        }
        
        return (SyncItem[])ret.toArray(new SyncItem[ret.size()]);
    }
           
    // --------------------------------------------------------- Private methods
    
    private SyncItemKey[] extractKeys(Collection syncItems) {
        SyncItemKey[] keys = new SyncItemKey[syncItems.size()];
        
        SyncItem syncItem = null;
        int j = 0;
        for(Iterator i = syncItems.iterator(); i.hasNext(); ++j) {
           syncItem = (SyncItem)i.next();
           
           keys[j] = syncItem.getKey();
        } // next i, j
        
        return keys;
    }
    
    private SyncItemKey[] extractKeys(SyncItem[] syncItems) {
        SyncItemKey[] keys = new SyncItemKey[syncItems.length];
        
        SyncItem syncItem = null;
        for(int j = 0; j<syncItems.length; ++j) {
           keys[j] = syncItems[j].getKey();
        } // next i, j
        
        return keys;
    } 
    
    /**
     * Look for LO and handle it accordingly.
     */
    private void handleLO() {
        SyncItem chunk = null;
        int p = -1;
        
        if (lo != null) {
            // We are reading a LO
            p = newSyncItems.indexOf(lo);
            if (p >= 0) {
                chunk = (SyncItem)newSyncItems.get(p);
            } else {
                p = updatedSyncItems.indexOf(lo);
                if (p >= 0) {
                    chunk = (SyncItem)updatedSyncItems.get(p);
                }
            }
            
            if (chunk != null) {
                mergeLOChunk(chunk);
                
                switch (lo.getState()) {
                    case SyncItemState.PARTIAL:
                        //
                        // just a new chunk
                        //
                        break;
                        
                    case SyncItemState.NEW:
                        //
                        // Last chunk of an addition
                        //
                        newSyncItems.remove(chunk);
                        newSyncItems.add(lo);
                        lo = null;
                        break;
                    
                    case SyncItemState.UPDATED:
                    case SyncItemState.SYNCHRONIZED:
                        //
                        // Last chunk of an update
                        //
                        updatedSyncItems.remove(chunk);
                        updatedSyncItems.add(lo);
                        lo = null;
                        break;
                }
                    
            }
        } else {
            //
            // Look for the first chunk
            //
            
            Iterator i = newSyncItems.iterator();
            while(i.hasNext()) {
                chunk = (SyncItem)i.next();
                
                if (chunk.getState() == SyncItemState.PARTIAL) {
                    lo = newSyncItem(chunk);
                    lo.setState(SyncItemState.NEW);
                    break;
                }
            }
            
            i = updatedSyncItems.iterator();
            while(i.hasNext()) {
                chunk = (SyncItem)i.next();
                
                if (chunk.getState() == SyncItemState.PARTIAL) {
                    lo = newSyncItem(chunk);
                    lo.setState(SyncItemState.UPDATED);
                    break;
                }
            }
        }
    }
    
    /**
     * Merges the given <code>chunk</code> with the existing <code>lo</code>
     * item.
     *
     * @param cunk the chunk to merge
     *
     */
    private void mergeLOChunk(SyncItem chunk) {
        if (lo == null) {
            // 
            // we should never get here!
            //
            return;
        }
        
        byte[] loData = (byte[])lo.getPropertyValue(lo.PROPERTY_BINARY_CONTENT);
        byte[] chunkData = (byte[])chunk.getPropertyValue(lo.PROPERTY_BINARY_CONTENT);
        
        byte[] allData = new byte[loData.length + chunkData.length];
        
        System.arraycopy(loData, 0, allData, 0, loData.length);
        System.arraycopy(chunkData, 0, allData, loData.length, chunkData.length);
        
        SyncProperty size = lo.getProperty(lo.PROPERTY_SIZE);
        lo.setProperties(chunk.getProperties());
        lo.setPropertyValue(lo.PROPERTY_BINARY_CONTENT, allData);
        lo.setProperty(size);
    }
    
    /**
     * Create a new <i>SyncItem</i> with the content of an existing one. 
     *
     * @param key
     * @param fromSyncItem
     */
    private SyncItem newSyncItem(SyncItem fromSyncItem) {
        SyncItemKey mappedKey = fromSyncItem.getMappedKey();
        
        String mappedKeyValue = (mappedKey == null)
                              ? null
                              : mappedKey.getKeyAsString()
                              ;
        SyncItemImpl syncItem 
            = new SyncItemImpl(fromSyncItem.getSyncSource()          ,
                               fromSyncItem.getKey().getKeyAsString(),
                               mappedKeyValue                        ,
                               fromSyncItem.getState()               );
        
        syncItem.setProperties(fromSyncItem.getProperties());
        
        return syncItem;       
    }
}