/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.framework.server.session;

/**
 * This class represents the handler for a SyncML session. It coordinates and
 * handles the packages and messages as dictated by the protocol.
 * <p>
 * The entry point is <i>processMessage()</i>, which determines which message is
 * expected and what processing has to be done (depending on the value of
 * <i>currentState</i>).
 * <p>
 * In the current implementation separate initialization is required. 
 * <p>
 * <i>SessionHandler</i> makes use of a <i>SyncEngine</i> for all 
 * tasks not related to the handling of the protocol.
 * See <i>sync4j.framework.engine.SyncEngine</i> for more information.
 *
 * @see sync4j.framework.engine.SyncEngine
 *
 * @author  Stefano Fornari @ Funambol
 * @version $Id: SessionHandler.java,v 1.9.4.2 2005/01/26 16:55:05 stefano_fornari Exp $
 */

import sync4j.framework.core.*;
import sync4j.framework.protocol.*;
import sync4j.framework.server.error.InvalidCredentialsException;
import sync4j.framework.engine.SyncEngine;
import sync4j.framework.engine.pipeline.MessageProcessingContext;
import sync4j.framework.engine.MessageSizeCalculator;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.server.error.ServerFailureException;

/**
 * This interface represent the functionality a session object must provide.
 * Different implementations can provide their own policy, but they have to be
 * compliant at least to this contract.
 */
public interface SessionHandler {

    // --------------------------------------------------------------- Constants

    public static final int STATE_START            = 0x0000;
    public static final int STATE_END              = 0x0001;
    public static final int STATE_PKG1_RECEIVING   = 0x0002;
    public static final int STATE_PKG1_RECEIVED    = 0x0003;
    public static final int STATE_PKG3_RECEIVING   = 0x0004;
    public static final int STATE_PKG3_RECEIVED    = 0x0005;
    public static final int STATE_SESSION_ABORTED  = 0xFF01;
    public static final int STATE_ERROR            = 0xFFFF;


    // ----------------------------------------------------------- publi methods


    public String getSessionId();
    
    /**
     * The session a new session?
     *
     * @return true if the session is new, false otherwise
     */
    public boolean isNew();
    
    public void setNew(boolean newSession);
       
    /**
     * Returns true if the sessione has been authenticated.
     *
     * @return true if the sessione has been authenticated, false otherwise
     */
    public boolean isAuthenticated();
    
    /**
     * Processes the given message. See the class description for more 
     * information.
     *
     * @param message the message to be processed
     *
     * @return the response message
     *
     * @throws ProtocolException in case of a protocol error
     * @throws InvalidCredentialsException in case of invalid credentials
     * @throws ServerFailureException in case of an internal server error
     */
    public SyncML processMessage(SyncML message, MessageProcessingContext context)
    throws ProtocolException, InvalidCredentialsException, ServerFailureException;

    /**
     * Processes an error condition. This method is called when the error is
     * is not fatal and is manageable at a protocol/session level so that a
     * well formed SyncML message can be returned.
     * <p>
     * Note that the offending message <i>msg</i> cannot be null, meaning that
     * at least the incoming message was a SyncML message. In this context,
     * <i>RepresentationException</i>s are excluded.
     *
     * @param the offending message - NOT NULL
     * @param the exception representing the error condition - NOT NULL
     *
     * @throws sync4j.framework.core.Sync4jException in case of unexpected errors
     *
     * @return the response message
     */
    public SyncML processError(SyncML msg, Throwable error)
    throws Sync4jException;

    /**
     * Called by the <i>SessionManager</i> when the session is expired.
     * It logs out the credential and release aquired resources.
     */
    public void expire();
    
    /**
     * Gets the creation timestamp of the session
     */
    public long getCreationTimestamp();
    
    /**
     * Indicates that the current session successfully ended. The synch-
     * ronization can be finalized and committed.
     */
    public void commit();
    
    /**
     * Indicates that the current session must be forced to end without 
     * commiting the work (note that this does not automatically mean that the
     * synchronization is rolled back).
     *
     * @param statusCode the reason for the aborting
     */
    public void abort(int statusCode);

    /**
     * Gets the current state
     *
     * @return the current state
     */
    public int getCurrentState();
    
    /**
     * Set the size calculator to use to calculate messages size.
     *
     * @param calculator the calculator
     */
    public void setSizeCalculator(MessageSizeCalculator calculator);

}

