/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.admin.ws;

import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.framework.core.*;
import sync4j.framework.config.Configuration;
import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.config.LoggingConfiguration;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.server.error.*;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.server.Sync4jUser;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.server.Sync4jModule;
import sync4j.framework.server.store.Clause;
import sync4j.framework.tools.beans.BeanFactory;

import sync4j.server.admin.AdminManager;
import sync4j.server.admin.AdminException;

/**
 * This is the session enterprise java bean that handles the administration
 * console. It is designed to be a stateless session bean.
 * <p>
 *  This server accepts the requests addressed to the hostname
 *  indicated by the configuration property pointed by {CONFIG_SERVER_URI} (see
 *  Sync4j.properties).
 * <p>
 * AdminBean uses the system property sync4j.configpath to set the base of the
 * config path.
 *
 * @author  Luigia Fassina @ Funambol
 *
 * @version $Id: AdminWS.java,v 1.1.2.6 2005/01/20 16:09:55 luigiafassina Exp $
 *
 */
public class AdminWS {
    // ------------------------------------------------------- Private constants
    
    // ------------------------------------------------------------ Private data
    
    private AdminManager admin;
    
    private transient Logger log = Sync4jLogger.getLogger();
    
    // ------------------------------------------------------------ Constructors
    
    /**
     * Creates a new AdminWS instance. This will initialize <i>admin</i>
     *
     * @throws ServerException if the AdminManager cannot be created.
     *
     */
    public AdminWS() {
        admin = new AdminManager();
    }
        
    // ---------------------------------------------------------- Public methods
    
    /**
     * Read the list of roles available.
     *
     * @return names of roles available
     *
     * @throws ServerException
     * @throws AdminException
     */
    public String[] getRoles()
    throws ServerException, AdminException {
        return admin.getRoles();
    }
    
    /**
     * Read all users that satisfy the parameter of search.
     *
     * @param clause array of conditions for the query
     *
     * @return array of Sync4jUser
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jUser[] getUsers(String clause)
    throws ServerException, AdminException {
        try {
            return admin.getUsers((Clause)BeanFactory.unmarshal(clause));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "getUsers", e);
            throw new ServerException(e.getMessage(), e);
        }
    }
    
    /**
     * Adds a new user and the assigned role to it
     *
     * @param user the user to add
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void addUser(Sync4jUser user)
    throws ServerException, AdminException {
        admin.addUser(user);
    }
    
    /**
     * Update the information of the specific user
     *
     * @param user the user with informations updated
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setUser(Sync4jUser user)
    throws ServerException, AdminException {
        admin.setUser(user);
    }
    
    /**
     * Delete the user
     *
     * @param userName the name of user to delete
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deleteUser(String userName)
    throws ServerException, AdminException {
        admin.deleteUser(userName);
    }
    
    /**
     * Count the number of users that satisfy the specif clauses.
     *
     * @param clause the conditions of the search
     *
     * @return number of users
     *
     * @throws ServerException
     * @throws AdminException
     */
    public int countUsers(String clause)
    throws ServerException, AdminException {
        try {
            return admin.countUsers((Clause)BeanFactory.unmarshal(clause));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "countUsers", e);
            throw new ServerException(e.getMessage(), e);
        }
    }
    
    /**
     * Read a list of device that satisfy the specific conditions.
     *
     * @param clauses the conditions foe search informations.
     *
     * @return array of device
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jDevice[] getDevices(String clauses)
    throws ServerException, AdminException {
        try {
            return admin.getDevices((Clause)BeanFactory.unmarshal(clauses));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "getDevices", e);
            throw new ServerException(e.getMessage(), e);
        }   
    }
    
    /**
     * Adds a new device.
     *
     * @param d the new device
     *
     * @return the device id
     *
     * @throws ServerException
     * @throws AdminException
     */
    public String addDevice(Sync4jDevice d)
    throws ServerException, AdminException {
        return admin.addDevice(d);
    }
    
    /**
     * Update the information of specific device.
     *
     * @param d the device to update
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setDevice(Sync4jDevice d)
    throws ServerException, AdminException {
        admin.setDevice(d);
    }
    
    /**
     * Delete the specific device.
     *
     * @param deviceId the device id that identifies the device
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deleteDevice(String deviceId)
    throws ServerException, AdminException{
        admin.deleteDevice(deviceId);
    }
    
    /**
     * Count the number of device that satisfy the specific clauses.
     *
     * @param clauses the specific conditions for search device
     *
     * @return the number of device finds
     *
     * @throws ServerException
     * @throws AdminException
     */
    public int countDevices(String clauses)
    throws ServerException, AdminException {
        try {
            return admin.countDevices((Clause)BeanFactory.unmarshal(clauses));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "countDevices", e);
            throw new ServerException(e.getMessage(), e);
        }  
    }
    
    /**
     * Read all principal that satisfy the clauses.
     *
     * @param clauses the specific conditions for search principal
     *
     * @return array of principal
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jPrincipal[] getPrincipals(String clauses)
    throws ServerException, AdminException {        
        try {
            return admin.getPrincipals((Clause)BeanFactory.unmarshal(clauses));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "getPrincipals", e);
            throw new ServerException(e.getMessage(), e);
        } 
    }
    
    /**
     * Adds a new principal.
     *
     * @param p the new principal
     *
     * @return the principal id
     *
     * @throws ServerException
     * @throws AdminException
     */
    public String addPrincipal(Sync4jPrincipal p)
    throws ServerException, AdminException {
        return admin.addPrincipal(p);
    }
    
    /**
     * Delete the specific principal.
     *
     * @param principalId the principal id that identifies the principal
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deletePrincipal(String principalId)
    throws ServerException, AdminException {
        admin.deletePrincipal(principalId);
    }
    
    /**
     * Count the number of principal that satisfy the specific clauses.
     *
     * @param clauses the specific conditions for search principal
     *
     * @return the number of principal finds
     *
     * @throws ServerException
     * @throws AdminException
     */
    public int countPrincipals(String clauses)
    throws ServerException, AdminException {
        try {
            return admin.countPrincipals((Clause)BeanFactory.unmarshal(clauses));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "countPrincipals", e);
            throw new ServerException(e.getMessage(), e);
        }
    }
    
    /**
     * Read all modules information
     *
     * @return an array with the modules information (empty if no objects are found)
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jModule[] getModulesName()
    throws ServerException, AdminException {
        return admin.getModulesName();
    }
    
    /**
     * Read the information of the specific module
     *
     * @param moduleId the module id that identifies the module to search
     *
     * @return the relative information to the module
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jModule getModule2(String moduleId)
    throws ServerException, AdminException {
        return admin.getModule(moduleId);
    }
    
    public String getModule(String moduleId)
    throws ServerException, AdminException {
        try {
            return BeanFactory.marshal(admin.getModule(moduleId));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "getModule", e);
            throw new ServerException(e.getMessage(), e);
        }  
    }
        
    /**
     * Adds a new source into datastore and create a relative file xml with configuration.
     * The source must have a defined source type.
     * The source type must refer to a connector.
     * The connector must refer to a module.
     *
     * @param moduleId the module id
     * @param connectorId the connector id
     * @param sourceTypeId the source type id
     * @param source the information of the new source
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void addSource(String moduleId, String connectorId, String sourceTypeId, String source)
    throws ServerException, AdminException {
        try {
            admin.addSource(moduleId, connectorId, sourceTypeId, (SyncSource)BeanFactory.unmarshal(source));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "addSource", e);
            throw new ServerException(e.getMessage(), e);
        }
        
    }
    
    /**
     * Update a specific source into datastore and create a relative file xml with configuration.
     * The source must have a defined source type.
     * The source type must refer to a connector.
     * The connector must refer to a module.
     *
     * @param moduleId the module id
     * @param connectorId the connector id
     * @param sourceTypeId the source type id
     * @param source the information of the new source
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setSource(String moduleId, String connectorId, String sourceTypeId, String source)
    throws ServerException, AdminException{    
        try {
            admin.setSource(moduleId, connectorId, sourceTypeId, (SyncSource)BeanFactory.unmarshal(source));
        } catch (Exception e) {
            log.severe("Server error: " + e.getMessage());
            log.throwing(getClass().getName(), "setSource", e);
            throw new ServerException(e.getMessage(), e);
        }
    }
    
    /**
     * Delete a specific source and the relative file of configuration.
     *
     * @param sourceUri the uri that identifies the source
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deleteSource(String sourceUri)
    throws ServerException, AdminException{
        admin.deleteSource(sourceUri);
    }
    
    /**
     * Returns the server logging configuration
     *
     * @return the server logging configuration as a <i>LoggingConfiguration</i>
     *         object.
     *
     * @throws ServerException, AdminException
     */
    public LoggingConfiguration getLoggingConfiguration() 
    throws ServerException, AdminException {
        return admin.getLoggingConfiguration();
    }    
    
    /**
     * Saves and apply the new logging configuration
     *
     * @param config the new logging configuration
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setLoggingConfiguration(LoggingConfiguration config) 
    throws ServerException, AdminException {
        admin.setLoggingConfiguration(config);
    }    
    
    
    /**
     * This method is only used to check credentials. Being this WS stateless, 
     * its methods are always called on request and each of them has to pass
     * authentication. However, for instance the SyncAdmin, has a login panel
     * that is displayed before any access to one of the other WS methods. In
     * order to provide to the user an immediate feedback, the SyncAdmin (or any
     * other client can just call <i>login()</i>.
     *
     * @param username administrator username
     */
    public void login(String username) {
        if (log.isLoggable(Level.INFO)) {
            log.info("New administrative session for " + username);
        }
    }
    
    /**
     * Returns the server version.
     *
     * @return the server version
     *
     */
    public String getServerVersion(String version) {
        return admin.getServerVersion();
    }
    
    // --------------------------------------------------------- Private methods
}