/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.engine;

import java.util.*;
import java.util.Map;
import java.util.logging.Logger;
import java.util.logging.Level;

import java.security.Principal;
import java.sql.Timestamp;

import sync4j.framework.engine.*;
import sync4j.framework.engine.source.CommittableSyncSource;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceInfo;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.security.Officer;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.tools.beans.BeanFactory;
import sync4j.framework.core.*;
import sync4j.framework.database.Database;
import sync4j.framework.protocol.CommandIdGenerator;
import sync4j.framework.config.Configuration;
import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.server.*;
import sync4j.framework.server.session.SyncSourceState;
import sync4j.framework.server.store.*;
import sync4j.framework.tools.MD5;
import sync4j.framework.tools.Base64;

/**
 * This is the Sync4j implementation of the synchronization engine.
 *
 * LOG NAME: sync4j.engine
 *
 * @author  Stefano Fornari @ Funambol
 * @version $Id: Sync4jEngine.java,v 1.52.4.9 2005/01/27 12:21:54 stefano_fornari Exp $
 */
public class Sync4jEngine
extends AbstractSyncEngine
implements java.io.Serializable, ConfigurationConstants {

    // --------------------------------------------------------------- Constants

    public static final String LOG_NAME = "engine";

    // ------------------------------------------------------------ Private data

    /**
     * The configuration properties
     */
    private Configuration configuration = null;

    public Configuration getConfiguration() {
        return configuration;
    }

    /**
     * The security officer
     */
    private Officer officer = null;

    public Officer getOfficer() {
        return this.officer;
    }

    /**
     * The underlying persistent store
     */
    private PersistentStore store = null;

    /** Getter for property store.
     * @return Value of property store.
     *
     */
    public PersistentStore getStore() {
        return store;
    }

    /** Setter for property store.
     * @param store New value of property store.
     *
     */
    public void setStore(PersistentStore store) {
        this.store = store;
    }

    /**
     * The sources this engine deals with
     */
    private ArrayList clientSources = new ArrayList();
    private HashMap serverSources = new HashMap();

    /**
     * The modification objects of the last synchronization
     */
    private HashMap operations = null;

    /**
     * The status of the last synchronization modifications
     */
    private ArrayList status = null;
    
    /**
     * Logger
     */
    protected transient Logger log = null;

    /**
     * Sets the SyncML Protocol version. It is used to create the server
     * credential with MD5 authentication.
     */
    private String syncMLVerProto = null;
    public void setSyncMLVerProto(String syncMLVerProto) {
        this.syncMLVerProto = syncMLVerProto;
    }

    private boolean isLastMessage  = false;
    public void setLastMessage(final boolean isLastMessage) {
        this.isLastMessage = isLastMessage;
    }
    
    public boolean isLastMessage() {
        return this.isLastMessage;
    }
    
    /**
     * The SyncSoure states
     */
    private SyncSourceState syncSourceState;
    
    /**
     * The existing LUID-GUID mapping. It is used and modify by sync()
     */
    private Map clientMappings;
    
    // ------------------------------------------------------------ Constructors

    /**
     * To allow deserializazion of subclasses.
     */
    protected  Sync4jEngine() {
        syncSourceState = new SyncSourceState();
    }

    /**
     * Creates a new instance of Sync4jEngine. <p>
     * NOTE: This is a sample implementation that deals with a single source on
     * the file system.
     *
     * @throws ConfigurationException a runtime exception in case of misconfiguration
     */
    public Sync4jEngine(Configuration configuration) {
        this();
        
        log = Sync4jLogger.getLogger(LOG_NAME);

        this.configuration = configuration;

        //
        // Set the underlying persistent store
        //
        store = configuration.getStore();

        //
        // Set the security officer
        //
        officer = configuration.getOfficer();

        //
        // Set the SyncStrategy object to use for comparisons
        //
        SyncStrategy strategy = configuration.getStrategy();

        setStrategy(strategy);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Engine configuration:");
            log.finest("store: "    + store   );
            log.finest("officer: "  + officer );
            log.finest("strategy: " + strategy);
        }
    }

    // ------------------------------------------------ Configuration properties

     /**
     * The id generator
     */
    private CommandIdGenerator cmdIdGenerator = null;

    /**
     * Set the id generator for commands
     *
     * @param cmdIdGenerator the id generator
     */
    public void setCommandIdGenerator(CommandIdGenerator cmdIdGenerator) {
        this.cmdIdGenerator = cmdIdGenerator;
    }


    /**
     * Return the id generator
     *
     * @return the id generator
     */
    public CommandIdGenerator getCommandIdGenerator() {
        return cmdIdGenerator;
    }

    // ------------------------------------------------------ Runtime properties

    /**
     * The database to be synchronized
     */
    private HashMap dbs = new HashMap();

    public void setDbs(HashMap dbs) {
        if (this.dbs != null) {
            this.dbs.clear();
        } else {
            this.dbs = new HashMap(dbs.size());
        }
        this.dbs.putAll(dbs);
    }

    public void setDbs(Database[] dbs) {
        if (this.dbs == null) {
            this.dbs = new HashMap(dbs.length);
        }

        for (int i=0; ((dbs != null) && (i<dbs.length)); ++i) {
            this.dbs.put(dbs[i].getName(), dbs[i]);
        }
    }

    public Database[] getDbs() {
        if (this.dbs == null) {
            return new Database[0];
        }

        Iterator i = this.dbs.values().iterator();

        Database[] ret = new Database[this.dbs.size()];

        int j = 0;
        while(i.hasNext()) {
            ret[j++] = (Database)i.next();
        }

        return ret;
    }
    
    /**
     * Timestamp of the beginning of this Sync
     */
    private Timestamp syncTimestamp;
    
    /**
     * Setter for syncTimestamp
     *
     * @param syncTimestamp new value
     */
    public void setSyncTimestamp(Date syncTimestamp) {
        //
        // Just for now: we have to change everything to Date
        //
        this.syncTimestamp = new Timestamp(syncTimestamp.getTime());
    }


    // ---------------------------------------------------------- Public methods

    /**
     * Fires and manages the synchronization process.
     *
     * @param principal the principal who is requesting the sync
     *
     * @throws Sync4jException in case of error
     */
    public void sync(final Sync4jPrincipal principal)
    throws Sync4jException {
        if (log.isLoggable(Level.INFO)) {
            log.info("Starting synchronization ...");
        }

        SyncStrategy syncStrategy = getStrategy();

        SyncSource clientSource = null, serverSource = null;
        Database   db           = null;

        //
        // Now process clientSources
        //

        operations = new HashMap();
        status     = new ArrayList();

        String   uri = null;
        ListIterator c = clientSources.listIterator();
        while(c.hasNext()) {
            clientSource = (SyncSource)c.next();
            uri          = clientSource.getSourceURI();
            
            if (log.isLoggable(Level.FINEST)) {
                log.finest( "SyncSource state of '"
                          + uri
                          + "' is "
                          + syncSourceState.getStateName(uri)
                          );
            }
            
            //
            // If the source state is the error state, skip it
            //
            if (syncSourceState.isError(uri)) {
                continue;
            }
            
            serverSource = getServerSource(uri);
            db           = (Database)dbs.get(uri);

            SyncSource[] sources = new SyncSource[] {
                serverSource, clientSource
            };

            //
            // I can't remember why this is done. If somebody reads this and 
            // knows the reason, please UPDATE this comment!
            //
            Sync4jPrincipal p = (Sync4jPrincipal)db.getPrincipal();
            p.setId(principal.getId());
            
            //
            // if the source is in the CONFIGURED state, call beginSync(). If 
            // beginSync() fails, skip 
            //
            if (syncSourceState.isReady(uri)) {
                try {
                    serverSource.beginSync(p, db.getMethod());
                    syncSourceState.moveToSyncing(uri);
                } catch (SyncSourceException e) {
                    syncSourceState.moveToError(uri);
                    setDBError(db, e);
                    continue;
                }
            }

            //
            // Note: in the case of fast sync, we do prepareSync and sync as 
            // soon as we can. If instead the slow sync is the case, we do the
            // sync preparation and the sync only at the end of the process, 
            // when all items from the client have been sent. 
            // Plus, in the case of a slow sync, the status commands in response
            // of a modification command is usually 200, since no real handling
            // is done.
            //
            if (   (db.getMethod()     == AlertCode.SLOW)
                || (db.getMethod()     == AlertCode.REFRESH_FROM_SERVER)
                || (db.getStatusCode() == StatusCode.REFRESH_REQUIRED)) {
                    
                try {
                    operations.put(uri, syncStrategy.prepareSlowSync(sources, p, syncTimestamp, isLastMessage));

                    status.addAll(
                        Arrays.asList(
                            syncStrategy.sync((SyncOperation[])operations.get(uri))
                        )
                    );
                } catch (SyncException e) {
                    syncSourceState.moveToError(uri);
                    setDBError(db, e);
                    continue;
                }  
            } else {
                //
                // Read the last timestamp from the persistent store, than
                // prepare for fast synchronization
                //
                LastTimestamp last = new LastTimestamp(
                                         p.getId()    ,
                                         db.getName()
                                     );
                try {
                    store.read(last);
                } catch (PersistentStoreException e) {
                    throw new Sync4jException("Error reading last timestamp", e);
                }
                
                Timestamp since = new Timestamp(last.start);

                try {
                    operations.put(uri, syncStrategy.prepareFastSync(sources, p, since, isLastMessage));
                    //
                    // Now synchronize the sources
                    //
                    status.addAll(
                        Arrays.asList(
                            syncStrategy.sync((SyncOperation[])operations.get(uri))
                        )
                    );
                } catch (SyncException e) {
                    syncSourceState.moveToError(uri);
                    setDBError(db, e);
                    continue;
                }  
            }
                        
            //
            // Let's try to commit the sync now!
            //
            if (serverSource instanceof CommittableSyncSource) {
                try {
                    ((CommittableSyncSource)serverSource).commitSync();
                    syncSourceState.moveToCommitted(uri);
                } catch (SyncSourceException e) {
                    syncSourceState.moveToError(uri);
                    setDBError(db, e);
                    continue;
                }
            } else {
                //
                // If the sync source is not committable, just make it move to
                // the committed state since commit is implicit.
                //
                syncSourceState.moveToCommitted(uri);
            }
            
            //
            // After processing, the database status code is set to OK
            //
            db.setStatusCode(StatusCode.OK);
            
        }  // next i (client source)
    
        //
        // Call endSync()
        //
        if (isLastMessage) {
            while (c.hasPrevious()) {
                clientSource = (SyncSource)c.previous();
                uri          = clientSource.getSourceURI();
                serverSource = getServerSource(uri);
                
                try {
                    serverSource.endSync(principal);
                } catch (SyncException e) {
                    log.throwing(getClass().getName(), "sync", e);
                    if (log.isLoggable(Level.SEVERE)) {
                        log.severe(e.getMessage());
                    }
                }
            }
        }

        log.info("Ending synchronization ...");

        try {
            syncStrategy.endSync();
        } catch (SyncException e) {
            log.throwing(getClass().getName(), "sync", e);
            if (log.isLoggable(Level.SEVERE)) {
                log.severe(e.getMessage());
            }
        }
        
    }

    /**
     * Returns the operations of the last synchronization
     *
     * @param uri the URI of the source the operations are applied to
     *
     * @return the operations of the last synchronization
     */
    public SyncOperation[] getSyncOperations(String uri) {
        return (SyncOperation[])operations.get(uri);
    }
    
    /**
     * Resets the operations for the given source.
     *
     * @param uri source uri
     */
    public void resetSyncOperations(String uri) {
        operations.put(uri, new SyncOperation[0]);
    }

    /**
     * Returns the operation status of the operations executed in the last
     * synchronization
     *
     * @param msgId the id of the SyncML message
     *
     * @return the operations status of the operations executed in the last
               synchronization
     */
    public Status[] getModificationsStatusCommands(String msgId) {

        if ((status == null) || (status.size() == 0)) {
            return new Status[0];
        }

        SyncOperationStatus[] operationStatus =
            (SyncOperationStatus[])status.toArray(new SyncOperationStatus[status.size()]);
        
        return EngineHelper.generateStatusCommands(
            operationStatus,
            msgId,
            cmdIdGenerator
        );
    }

    /**
     * Adds a new client source to the list of the sources the engine deals with.
     *
     * @param source the source to be added
     */
    public void addClientSource(SyncSource source) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("adding " + source);
        }

        clientSources.add(source);
        
        syncSourceState.moveToReady(source.getSourceURI());
    }

    /**
     * Returns the client sources
     *
     * @return the client sources
     */
    public List getClientSources() {
        return clientSources;
    }

    /**
     * Returns the client source with the given name if there is any.
     *
     * @param name the source name
     *
     * @return the client source with the given name if there is any, null otherwise
     */
    public SyncSource getClientSource(String name) {
        Iterator i = clientSources.iterator();

        SyncSource s = null;
        while (i.hasNext()) {
            s = (SyncSource)i.next();

            if (s.getSourceURI().equals(name)) {
                return s;
            }
        }

        return null;
    }
    
    /**
     * Returns the status of the give client source (as stored in the 
     * corresponding database object).
     *
     * @param uri the source uri
     *
     * @return the status of the client source (as stored in the corresponding
     * database object)
     */
    public int getClientSourceStatus(String uri) {
        Database db = (Database)dbs.get(uri);
        
        return db.getStatusCode();
    }
    
    /**
     * Returns the status message of the given client source (as stored in the 
     * corresponding database object).
     *
     * @param uri the source uri
     *
     * @return the status message of the client source (as stored in the corresponding
     * database object)
     */
    public String getClientStatusMessage(String uri) {
        Database db = (Database)dbs.get(uri);
        
        return db.getStatusMessage();
    }

    /**
     * Returns the server source with the given name if there is any.
     *
     * @param name the source name
     *
     * @return the server source with the given name if there is any, null otherwise
     */
    public SyncSource getServerSource(String name) {
        if (serverSources.containsKey(name)) {
            return (SyncSource)serverSources.get(name);
        }
        
        Sync4jSource s = new Sync4jSource(name);

        try {
            store.read(s);
        } catch (PersistentStoreException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe( "Error reading source configuration for source "
                          + name
                          + ": "
                          + e.getMessage()
                          );
            }

            log.throwing(getClass().getName(), "getServerSource", e);
        }

        try {
            SyncSource syncSource = (SyncSource)configuration.getBeanInstanceByName(s.getConfig());
            serverSources.put(name, syncSource);
            return syncSource;
            
        } catch (Exception e){
            String msg = "Unable to create sync source "
                       + s
                       + ": "
                       + e.getMessage();

            log.severe(msg);
            log.throwing(getClass().getName(), "getServerSource", e);
        }

        return null;
    }

    /**
     * Return the <i>DataStore</i> object corresponding to the given
     * <i>Database</i> object.
     *
     * @param database the database to convert
     *
     * @return the <i>DataStore</i> object corresponding to the given
     *         <i>Database</i> object.
     *
     */
    public DataStore databaseToDataStore(Database db) {
        ContentTypeInfo contentTypeInfo
                         = new ContentTypeInfo(db.getType(), "-" /* version */);

        return new DataStore(
                    new SourceRef(db.getSource())                ,
                    null                                         ,
                    -1                                           ,
                    contentTypeInfo                              ,
                    new ContentTypeInfo[] { contentTypeInfo }    ,
                    contentTypeInfo                              ,
                    new ContentTypeInfo[] { contentTypeInfo }    ,
                    new DSMem(false, -1, -1)                     ,
                    new SyncCap(new SyncType[] { SyncType.SLOW })
        );
    }

    /**
     * Returns the content type capabilities of this sync engine.
     *
     * @return the content type capabilities of this sync engine.
     */
    public CTCap[] getContentTypeCapabilities() {
        return new CTCap[0];
    }

    /**
     * Returns the extensions of this sync engine.
     *
     * @return the extensions of this sync engine.
     */
    public Ext[] getExtensions() {
        return new Ext[0];
    }

    /**
     * Returns the data store capabilities of the sync sources to synchronize.
     *
     * @return the data store capabilities of the sync sources to synchronize.
     */
    public DataStore[] getDatastores() {
        DataStore[] ds = null;
        ArrayList al = new ArrayList();

        Database db  = null;
        String   uri = null;
        Iterator i = dbs.values().iterator();
        while(i.hasNext()) {
            db = (Database)i.next();
            uri = db.getName();
            
            if (!db.isOkStatusCode()) {
                continue;
            }

            SyncSource ss = getServerSource(uri);

            if ( ss != null) {
                al.add(EngineHelper.toDataStore(uri, ss.getInfo()));
            }
        }

        int size = al.size();
        if (size == 0) {
            ds = new DataStore[0];
        } else {
        ds = (DataStore[])al.toArray(new DataStore[size]);
        }


        return ds;
    }


    /**
     * Creates and returns a <i>DeviceInfo</i> object with the information
     * extracted from the configuration object.
     *
     * @param verDTD the version of the DTD to use to encode the capabilities
     *
     * @return the engine capabilities
     *
     * @throws ConfigurationException a runtime exception in case of misconfiguration
     */
    public DevInf getServerCapabilities(VerDTD verDTD) {
        DevInf devInf =
        new DevInf(
            verDTD,
            configuration.getStringValue(CFG_ENGINE_MANIFACTURER, "Sync4j"),
            configuration.getStringValue(CFG_ENGINE_MODELNAME,    "Sync4j"),
            configuration.getStringValue(CFG_ENGINE_OEM,          null    ),
            configuration.getStringValue(CFG_ENGINE_FWVERSION,    null    ),
            configuration.getStringValue(CFG_ENGINE_SWVERSION,    null    ),
            configuration.getStringValue(CFG_ENGINE_HWVERSION,    null    ),
            configuration.getStringValue(CFG_ENGINE_DEVICEID,     "sync4j"),
            configuration.getStringValue(CFG_ENGINE_DEVICETYPE,   "Server"),
            getDatastores(),
            getContentTypeCapabilities(),
            getExtensions(),
            false,
            false,
            false
        );
        return devInf;
    }

    /**
     * First of all, check the availablity and accessibility of the given
     * databases. The state of the given database will change as described below
     * (and in the same order):
     * <ul>
     *   <li>The database status code is set to <i>StatusCode.OK</i> if the
     *       database is accessible, <i>StatusCode.NOT_FOUND</i> if the database
     *       is not found and <i>StatusCode.FORBIDDEN</i> if the principal is
     *       not allowed to access the database.
     *   <li>If the currently set last anchor does not match the last tag
     *       retrieved from the database (DBMS) and the requested alert code is
     *       not a refresh, the synchronization method is set to
     *       <i>AlertCode.SLOW</i>
     *   <li>The database server sync anchor is set to the server-side sync anchor
     *
     * @param principal the principal the is requesting db preparation
     * @param dbs an array of <i>Database</i> objects - NULL
     * @param next the sync timestamp of the current synchronization
     *
     */
    public void prepareDatabases(Sync4jPrincipal principal,
                                 Database[]      dbs      ,
                                 SyncTimestamp   next     ) {
        for (int i=0; ((dbs != null) && (i < dbs.length)); ++i) {
            int statusCode = StatusCode.OK;

            if (!checkServerDatabase(dbs[i])) {
                statusCode = StatusCode.NOT_FOUND;
            } else if (!checkDatabasePermissions(dbs[i])) {
                statusCode = StatusCode.FORBIDDEN;
            }

            dbs[i].setStatusCode(statusCode);

            //
            // Now retrieve the last sync anchor
            //
            if (statusCode == StatusCode.OK) {
                LastTimestamp last = new LastTimestamp(
                    principal.getId(),
                    dbs[i].getName()
                );

                //
                // Into tagServer there is the Next anchor sent by client: now
                // this become a Last anchor of the server.
                // Into tagClient there is the Next anchor sent by server: now
                // this become a Last anchor of the client.
                // The server anchor has like Last the Next sent from the client
                // and like Next the Next generated at the start of sync.
                //
                try {
                    store.read(last);
                    dbs[i].setServerAnchor(new Anchor(last.tagClient, next.tagClient));
                } catch (NotFoundException e) {
                    //
                    // No prev last sync! Create a new anchor that won't match
                    //
                    last.tagServer = next.tagClient;
                    dbs[i].setServerAnchor(new Anchor(last.tagServer, next.tagClient));
                } catch(PersistentStoreException e) {
                    log.severe("Unable to retrieve timestamp from store");
                    log.throwing(getClass().getName(), "prepareDatabases", e);
                }

                if (  !(last.tagServer.equals(dbs[i].getAnchor().getLast()))
                   && (dbs[i].getMethod() != AlertCode.REFRESH_FROM_SERVER)) {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest( "Forcing slow sync for database "
                                + dbs[i].getName()
                                );
                        log.finest( "Server last: "
                                + last.tagServer
                                + "; client last: "
                                + dbs[i].getAnchor().getLast()
                                );
                    }
                    if (dbs[i].getMethod() != AlertCode.SLOW) {
                        dbs[i].setStatusCode(StatusCode.REFRESH_REQUIRED);
                    }
                    dbs[i].setMethod(AlertCode.SLOW);
                }
            }
        }
    }

    /**
     * Authenticates the given credential using <i>officer</i>.
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is authenticated, false otherwise
     */
    public boolean login(Cred credential) {
        return officer.authenticate(credential);
    }

    /**
     * Logs out the given credential using <i>officer</i>.
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is authenticated, false otherwise
     */
    public void logout(Cred credential) {
        officer.unAuthenticate(credential);
    }

    /**
     * Authorizes the given principal to access the given resource using
     * <i>officer</i>.
     *
     * @param principal the entity to be authorized
     * @param resource the name of the resource
     *
     * @return true if the principal is authorized, false otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return officer.authorize(principal, resource);
    }

    /**
     * Translates an array of <i>SyncOperation</i> objects to an array of
     * <i>(Add,Delete,Replace)Command</i> objects. Only client side operations
     * are translated.
     *
     * @param operations the operations to be translated
     * @param uri the source uri
     * @param slow is a slow sync?
     *
     * @return the sync4j commands corresponding to <i>operations</i>
     */
    public ItemizedCommand[] operationsToCommands(SyncOperation[]    operations,
                                                  String             uri       ,
                                                  boolean            slow      ) {
        ItemizedCommand[] commands = 
            EngineHelper.operationsToCommands( getMapping(uri),
                                               operations     ,
                                               uri            ,
                                               cmdIdGenerator );
        try {
            EngineHelper.updateClientMappings(clientMappings, operations, slow);
        } catch (Exception e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "updateClientMappings", e);
        }
        
        return commands;
    }


    /**
     * Updates the LUID-GUIDmapping for the client modifications (that is the
     * items directlry inserted or removed by the server).
     *
     * @param slowSync is a slow sync in progress?
     *
     */
    public void updateServerMappings(boolean slowSync) {
        SyncOperationStatus[] operationStatus =
            (SyncOperationStatus[])status.toArray(new SyncOperationStatus[status.size()]);
        
        try {
            EngineHelper.updateServerMappings(clientMappings, operationStatus, slowSync);
        } catch (Exception e) {
            log.throwing(getClass().getName(), "updateServerMappings", e);
          
            if (log.isLoggable(Level.SEVERE)) {
                log.severe(e.getMessage());
            }
        }
    }

    /**
     * Converts an array of <i>Item</i> objects belonging to the same
     * <i>SyncSource</i> into an array of <i>SyncItem</i> objects.
     * <p>
     * The <i>Item</i>s created are enriched with an additional property
     * called as in SyncItemHelper.PROPERTY_COMMAND, which is used to bound the
     * newly created object with the original command.
     *
     * @param syncSource the <i>SyncSource</i> items belong to - NOT NULL
     * @param items the <i>Item</i> objects
     * @param state the state of the item as one of the values defined in
     *              <i>SyncItemState</i>
     * @param the timestamp to assign to the last even on this item
     *
     *
     * @return an array of <i>SyncItem</i> objects
     */
    public SyncItem[] itemsToSyncItems(SyncSource          syncSource   ,
                                       ModificationCommand cmd          ,
                                       char                state        ,
                                       long                timestamp    ) {

        ClientMapping m = getMapping(syncSource.getSourceURI());
        return EngineHelper.itemsToSyncItems(m, syncSource, cmd, state, timestamp);
    }

    
        /**
     * Returns the server credentials in the format specified by the given Chal.
     *
     * @param chal the client Chal
     * @param device the device
     *
     * @return the server credentials
     */
    public Cred getServerCredentials(Chal chal, Sync4jDevice device) {
        String login = configuration.getStringValue(CFG_SERVER_ID);

        if (login == null || login.equals("")) {
            return null;
        }

        String pwd    = device.getServerPassword();

        String type   = chal.getType();
        String data   = login + ':' + pwd;

        if (Constants.AUTH_TYPE_BASIC.equals(type)) {
            data = new String(Base64.encode(data.getBytes()));
        } else if (Constants.AUTH_TYPE_MD5.equals(type)) {

            if (this.syncMLVerProto.indexOf("1.0") != -1) {
                //
                // Steps to follow:
                // 1) decode nonce sent by client
                // 2) b64(H(username:password:nonce))
                //
                byte[] serverNonce = chal.getNextNonce().getValue();
                String serverNonceDecode = null;
                if ((serverNonce == null) || (serverNonce.length == 0)) {
                    serverNonceDecode = "";
                } else {
                    serverNonceDecode = Constants.FORMAT_B64.equals(chal.getFormat())
                                      ? new String(Base64.decode(serverNonce))
                                      : new String(serverNonce)
                                      ;
                }
                String withNonce = new String(data.getBytes()) + ':' + serverNonceDecode;
                byte[] digestNonce = MD5.digest(withNonce.getBytes());
                data =  new String(Base64.encode(digestNonce));

            } else {
                //
                //Steps to follow:
                //1) H(username:password)
                //2) b64(H(username:password))
                //3) decode nonce sent by client
                //4) b64(H(username:password)):nonce
                //5) H(b64(H(username:password)):nonce)
                //6) b64(H(b64(H(username:password)):nonce))
                //
                byte[] serverNonce = chal.getNextNonce().getValue();
                byte[] digest = MD5.digest(data.getBytes());
                byte[] b64    = Base64.encode(digest);

                String serverNonceDecode = null;
                if ((serverNonce == null) || (serverNonce.length == 0)) {
                    serverNonceDecode = "";
                } else {
                    serverNonceDecode = Constants.FORMAT_B64.equals(chal.getFormat())
                                      ? new String(Base64.decode(serverNonce))
                                      : new String(serverNonce)
                                      ;
                }

                String withNonce = new String(b64) + ':' + serverNonceDecode;
                byte[] digestNonce = MD5.digest(withNonce.getBytes());

                data =  new String(Base64.encode(digestNonce));
            }
        }

        Meta m = new Meta();
        m.setType(type);
        m.setNextNonce(null);
        return new Cred(new Authentication(m, data));
    }

    /**
     * Read the given principal from the store. The principal must be
     * already configured with username and device. The current implementation
     * reads the following additional informatioin:
     * <ul>
     *  <li>principal id
     * </ul>
     *
     * @param principal the principal to be read
     *
     * @throws sync4j.framework.server.store.PersistentStoreException
     */
    public void readPrincipal(Sync4jPrincipal principal)
    throws PersistentStoreException {
        assert(principal               != null);
        assert(principal.getUsername() != null);
        assert(principal.getDeviceId() != null);

        getStore().read(principal);
    }

    /**
     * Read the Sync4jDevice identified by the given id from the store.
     *
     * @param deviceId the device of which reading the information
     *
     * @throws sync4j.framework.server.store.PersistentStoreException
     */
    public void readDevice(Sync4jDevice device)
    throws PersistentStoreException {
        assert(device               != null);
        assert(device.getDeviceId() != null);

        getStore().read(device);
    }

    /**
     * Stores the given Sync4jDevice.
     *
     * @param device the device
     *
     */
    public void storeDevice(Sync4jDevice device) {
        try{
            getStore().store(device);
        } catch (PersistentStoreException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error storing the device: " + e.getMessage());
                log.throwing(getClass().getName(), "storeDevice", e);
            }
        }
    }
    
    /**
     * Updates the given mapping for the given source
     *
     * @param uri source uri
     * @param GUID the GUID
     * @param LUID the LUID
     *
     */
    public void updateMapping(String uri, String guid, String luid) {
        getMapping(uri).updateMapping(guid, luid);
    }
    
    
    
    /**
     * Returns the GUID-LUID mappings associated to the given source URI. 
     * Mappings are cached internally so that they are read from the db only
     * the first time they are requested.
     * Note that in the case of a slow sync, the first time a mapping is 
     * requested the mapping must be reset so that old values will be removed.
     *
     * @param principal principal - this is used only when the mapping was not
     *        cached yet; after caching it can be null
     * @param uri database URI
     * @param slow is a slow sync being performed? 
     *
     * @return the LUID-GUID mapping associated to the give source uri
     */
    public ClientMapping getMapping(Sync4jPrincipal principal, 
                                    String          uri      ,
                                    boolean         slow     ) {
        if (clientMappings == null) {
            clientMappings = new HashMap();
        }
        
        ClientMapping ret = (ClientMapping)clientMappings.get(uri);
        
        if (ret == null) {
            try {
                PersistentStore ps = getStore();

                ret = new ClientMapping(principal, uri);

                ps.read(ret);

                if (slow) {
                    ret.clearMappings();
                    ps.store(ret);
                }
                
                clientMappings.put(uri, ret);
            } catch (PersistentStoreException e) {
                log.severe("Unable to read clientMappings from the persistent store");
                log.throwing(getClass().getName(), "getMapping", e);
            }
        }
        
        return ret;
    }
    
    /**
     * This is a shortcut for getMapping(null, uri, false). It can be used when
     * it is sure that a mapping was already retrieved from the db.
     *
     * @param uri the source uri
     *
     * @return the LUID-GUID mapping associated to the give source uri
     */
    public ClientMapping getMapping(String uri) {
        return getMapping(null, uri, false);
    }
    
    /**
     * Stores the LUIG-GUID mappings into the database
     *
     * @throws Sync4jException in case of database error
     */
    public void storeMappings() {
        if (clientMappings == null) {
            return;
        }

        try {
            PersistentStore ps = getStore();

            ClientMapping cm = null;
            Iterator i = clientMappings.keySet().iterator();
            while (i.hasNext()) {
                cm = (ClientMapping)clientMappings.get(i.next());
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Saving client mapping: " + cm);
                }
                ps.store(cm);
                
                cm.resetModifiedKeys();
            }
        } catch (PersistentStoreException e) {
            log.severe("Unable to save clientMappings to the persistent store");
            log.throwing(getClass().getName(), "storeMappings", e);
        }
    }

    
    // --------------------------------------------------------- Private methods

    /**

    /**
     * Checks if the given database is managed by this server.
     *
     * @param db the database to be checked
     *
     * @return true if the database is under control of this server, false otherwise
     *
     */
    private boolean checkServerDatabase(Database db) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest( "Checking if the database "
                      + db
                      + " is in the server database list."
                      );
        }

        Sync4jSource s = new Sync4jSource(db.getName());

        try {
            store.read(s); log.finest("Yes sir!");

            return true;
        } catch (NotFoundException e) {
            log.finest("Not found sir");
        } catch (Exception e) {
            String msg = "Unable to access the perstistent store "
                       + s
                       + ": "
                       + e.getMessage();

            log.severe(msg);
            log.throwing(getClass().getName(), "checkServerDatabase", e);
        }

        //
        // not found...
        //
        return false;
    }

    /**
     * Checks if the principal associated to the database is allowed to synchronize.<br>
     *
     * @param db the database to be checked
     *
     * @return true if the credential is allowed to access the database, false otherwise
     */
    private boolean checkDatabasePermissions(Database db) {
        return authorize(db.getPrincipal(), db.getName());
    }
    
    /**
     * Sets the database error accordingly to the given exception. 
     *
     * @param db the database
     * @param cause the throwable cause
     *
     */
    private void setDBError(Database db, SyncException cause) {
        db.setStatusCode   (cause.getStatusCode());
        db.setStatusMessage(cause.getMessage()   );
    }
    
    /**
     * Resets the clientMappings map.
     */
    private void resetClientMappings() {
        clientMappings = new HashMap();
    }
   
}



