/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.engine;

import java.util.Arrays;
import java.util.ArrayList;
import java.util.Date;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import java.util.logging.*;
import java.security.Principal;
import java.sql.Timestamp;

import sync4j.framework.core.ModificationCommand;
import sync4j.framework.core.StatusCode;
import sync4j.framework.core.Add;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.*;

import sync4j.server.engine.SyncItemHelper;
import sync4j.server.engine.Sync4jOperationStatusOK;
import sync4j.server.engine.Sync4jOperationStatusError;
import sync4j.server.engine.Sync4jOperationStatusConflict;

import org.apache.commons.collections.ListUtils;

/**
 * This class represents a synchronization process.
 *
 * The base concrete implementation of a synchronization strategy. It implements
 * the <i>ConcreteStrategy</i> partecipant of the Strategy Pattern.
 * <p>
 * The synchronization process is implemented in this class as follows:
 * <p>
 * Given a set of sources A, B, C, D, etc, the synchronization process takes place 
 * between two sources at a time: A is first synchronzed with B, then AB with
 * C, then ABC with D and so on. <br>
 * The synchronization process is divided in three phases: preparation, synchronization,
 * finalization. These phases correspond to the methdos prepareSync, sync and
 * endSync; they are call-back methods called sequentially by a driver object,
 * usually a SyncMethod object. In this way the SyncMethod object has a chance 
 * to communicate with the external world before carry on the process. 
 * For example a SyncMethod object can show to the user the items that are going
 * to be synchronized, so that the user can let the process carry on or stop it. <br>
 * prepareSync returns an array of SyncOperation, in which each element represents
 * a particular synchronization action, ie. create an item in the source A,
 * delete the item X from the source B, etc. Sometime it is not possible decide 
 * what to do, thus a SyncConflict operation is used. A conflict must be solved
 * by something external the synchronization process, for instance by a user
 * action. Below is a table of all possible situations.
 * <pre>
 *
 * | -------- | --- | --- | --- | --- | --- |
 * | Source A |     |     |     |     |     |
 * |    /     |  N  |  D  |  U  |  S  |  X  |     N : item new
 * | Source B |     |     |     |     |     |     D : item deleted
 * | -------- | --- | --- | --- | --- | --- |     U : item updated
 * |       N  |  O  |  O  |  O  |  O  |  B  |     S : item synchronized/unchanged
 * | -------- | --- | --- | --- | --- | --- |     X : item not existing
 * |       D  |  O  |  X  |  O  |  X  |  X  |     O : conflict
 * | -------- | --- | --- | --- | --- | --- |     A : item A replaces item B
 * |       U  |  O  |  O  |  O  |  B  |  B  |     B : item B replaces item A
 * | -------- | --- | --- | --- | --- | --- |
 * |       S  |  O  |  X  |  A  |  =  |  B  |
 * | -------- | --- | --- | --- | --- | --- |
 * |       X  |  A  |  X  |  A  |  A  |  X  |
 * | -------- | --- | --- | --- | --- | --- |
 *
 * </pre>
 *
 * @author Stefano Fornari @ Funambol
 * 
 * @version  $Id: Sync4jStrategy.java,v 1.26.2.1.2.14 2005/01/28 15:48:16 luigiafassina Exp $
 */
public class Sync4jStrategy implements SyncStrategy, java.io.Serializable {
    
    // --------------------------------------------------------------- Constants
    
    public static String LOG_NAME = "sync4j.framework.engine";
    
    private transient static Logger log = Logger.getLogger(LOG_NAME);
    
    // -------------------------------------------------------------- Properties
    
    /**
     * The synchronization source 
     */
    private transient SyncSource[] sources = null; 
    public SyncSource[] getSources(){ return sources; }

    public void setSources(SyncSource[] sources){ this.sources = sources; }

    /**
     * The process' name
     */
    public String getName(){ return name; }

    private String name;
    public void setName(String name){ this.name = name; }
    
    // ------------------------------------------------------------ Constructors
    
    public Sync4jStrategy() {
    }
    
    public Sync4jStrategy(SyncSource[] syncSources) {
        sources = syncSources;
    }
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Preparation for the synchronization. If <i>sources</i> is not null,
     * the preparation works on the given sources. Otherwise it works on the 
     * sources as they were set in the constructor.
     * Note that in the case of slow sync, the sync analysis to determine the 
     * new items on the server that must be returned to the client, we need to
     * wait for the last call to this method. This is used for protocols that 
     * supports a multi message session. At the contrary, client side 
     * modifications can be processed immediately.
     *
     *
     * @param sources the sources to be synchronized
     * @param nextSync the timestamp of the beginning of this synchronization
     * @param principal the entity for which the synchronization is required
     * @param last is this the last call to prepareSlowSync ?
     *
     * @return an array of SyncOperation, one for each SyncItem that must be
     *         created/updated/deleted or in conflict.
     *
     * @see sync4j.framework.engine.SyncStrategy
     */
    public SyncOperation[] prepareSlowSync(SyncSource[] sources  , 
                                           Principal    principal,
                                           Timestamp    nextSync ,
                                           boolean      last     ) 
    throws SyncException {
        
        if (sources != null) {
            this.sources = sources;
        }
        
        SyncItem[] Am, Bm;
        List syncOperations = new ArrayList();

        Logger log = Logger.getLogger(LOG_NAME);
        
        if (log.isLoggable(Level.INFO)) {
            log.info( "Preparing slow synchronization of source '"
                    + sources[1].getSourceURI()
                    + "' for " 
                    + principal 
                    + " ..."
                    );
            if (last) {
                log.info("Last call");
            } else {
                log.info("Not the last call");
            }
        }
        
        //
        // First process the items provided by the client 
        //
        Am = sources[1].getUpdatedSyncItems(principal, null);
        Bm = sources[0].getAllSyncItems(principal);
        
        //
        // Just in case...
        //
        if (Am == null) {
            Am = new SyncItem[0];
        }
        
        if (Bm == null) {
            Bm = new SyncItem[0];
        }
          
        //
        // If any item from the client has not a corresponding mapping, the 
        // server source must be queried for the item, since the client item 
        // could be the same of an existing server item. In this case, the old
        // unmapped item is replaced in Map by the newly mapped item.
        //
        ArrayList newlyMappedItems = new ArrayList();
        fixMappedItems(newlyMappedItems, Am, sources[0], principal);
        
        //
        // Because it is a slow sync, items state must be reset to S
        //
        EngineHelper.resetState(Am);
        EngineHelper.resetState(Bm);
        
        for (int i=0; i<Am.length; ++i) {
            if (!Am[i].isMapped()) {
                syncOperations.add(checkSyncOperation(principal, nextSync, Am[i], null));
            }
        }
        
        Iterator j = newlyMappedItems.iterator();
        while (j.hasNext()) {
            SyncItemMapping m = (SyncItemMapping)j.next();
            m.getSyncItemA().setState(SyncItemState.SYNCHRONIZED);
            m.getSyncItemB().setState(SyncItemState.SYNCHRONIZED);
            syncOperations.add(checkSyncOperation(principal, nextSync, m.getSyncItemA(), m.getSyncItemB()));
        }
        
        //
        // If it is the last call, we have also to consider what is in source
        // B but not in A
        //
        if (last) {
            //
            // In this case getAllSyncItems() returns the items in the datastore
            // reconstructed from the GUID-LUID mapping. In addition to these 
            // items we have the ones sent in the last message (retrieved 
            // calling getUpdatedSyncItems().
            //
            Am = sources[1].getAllSyncItems(principal);
            EngineHelper.resetState(Am);
            ArrayList all = new ArrayList(Arrays.asList(Am));
            all.addAll(Arrays.asList(sources[1].getUpdatedSyncItems(principal, null)));
            
            for (int i=0; i<Bm.length; ++i) {
                if (!all.contains(Bm[i])) {
                    syncOperations.add(checkSyncOperation(principal, nextSync, null, Bm[i]));
                }
            }
        }
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("operations: " + syncOperations);
        }
        
        if (log.isLoggable(Level.INFO)) {
            log.info("Preparation completed.");
        }
        
        return (SyncOperation[])syncOperations.toArray(new SyncOperationImpl[] {});
    }
    
    /**
     * Preparation for faset synchronization. If <i>sources</i> is not null,
     * the preparation operates on the given sources. Otherwise it works on the 
     * sources as they were set in the constructor.
     * <p>
     * Refer to the <a href="http://sync4j.sourceforge.net/web/project/architecture/sync4j-architecture.html">
     * architecture document</a> for details about the algoritm applied.
     *
     * @param sources the sources to be synchronized
     * @param principal the entity for which the synchronization is required
     * @param lastSync timestamp of the last synchronization
     * @param last is this the last call to prepareFastSync ?
     *
     * @return an array of SyncOperation, one for each SyncItem that must be
     *         created/updated/deleted or in conflict.
     *
     * @see sync4j.framework.engine.SyncStrategy
     */
    public SyncOperation[] prepareFastSync(SyncSource[] sources  , 
                                           Principal    principal,
                                           Timestamp    lastSync ,
                                           boolean      last     ) 
    throws SyncException {
        if (sources != null) {
            this.sources = sources;
        }
        
        // ---------------------------------------------------------------------
        
        List Am    ,  // items modified in A
             Bm    ,  // items modified in B
             AmBm  ,  // Am intersect Bm
             AAmBm ,  // items unmodified in A, but modified in B
             AmBBm ;  // items unmodified in B, but modified in A
        
        ArrayList syncOperations = null;

        SyncItem[] newA       = null, newB       = null,
                   updatedA   = null, updatedB   = null,
                   deletedA   = null, deletedB   = null;

        if (log.isLoggable(Level.INFO)) {
            log.info( "Preparing fast synchronization of source '"
                    + sources[1].getSourceURI()
                    + "' for " 
                    + principal
                    + " since "
                    + lastSync
                    + "..."
                    );
        }
        
        // ---------------------------------------------------------------------

        //
        // NOTE: simplified version - only two sources, the first one of which 
        // is the client
        //
        newA       = sources[1].getNewSyncItems      (principal, lastSync);
        updatedA   = sources[1].getUpdatedSyncItems  (principal, lastSync);
        deletedA   = sources[1].getDeletedSyncItems  (principal, lastSync);

        newB       = sources[0].getNewSyncItems      (principal, lastSync);
        updatedB   = sources[0].getUpdatedSyncItems  (principal, lastSync);
        deletedB   = sources[0].getDeletedSyncItems  (principal, lastSync);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("newA: " + Util.arrayToString(newA));
            log.finest("newB: " + Util.arrayToString(newB));

            log.finest("updatedA: " + Util.arrayToString(updatedA));
            log.finest("updatedB: " + Util.arrayToString(updatedB));

            log.finest("deletedA: " + Util.arrayToString(deletedA));
            log.finest("deletedB: " + Util.arrayToString(deletedB));
        }
        
        //
        // If any item from the client has not a corresponding mapping, the 
        // server source must be queried for the item, since the client item 
        // could be the same of an existing server item. In this case,  the old
        // unmapped item is replaced by the newly mapped item.
        //
        ArrayList newlyMappedItems = new ArrayList();
        
        fixMappedItems(newlyMappedItems, newA,     sources[0], principal);
        fixMappedItems(newlyMappedItems, updatedA, sources[0], principal);
        fixMappedItems(newlyMappedItems, deletedA, sources[0], principal);
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Newly mapped items: " + newlyMappedItems);
        }
        
        Am = new ArrayList(); Bm = new ArrayList();
        Am.addAll(Arrays.asList(newA      )); 
        Am.addAll(Arrays.asList(updatedA  ));
        Am.addAll(Arrays.asList(deletedA  ));
        Bm.addAll(Arrays.asList(newB      )); 
        Bm.addAll(Arrays.asList(updatedB  ));
        Bm.addAll(Arrays.asList(deletedB  ));
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Am: " + Am);
            log.finest("Bm: " + Bm);
            log.finest("Am-Bm: " + ListUtils.subtract(Am, Bm));
            log.finest("Bm-Am: " + ListUtils.subtract(Bm, Am));
        }
        
        //
        // Now calculate subsets: AmBm, AmBBm, AAmBm
        //
        AmBm  = EngineHelper.intersect(Am, Bm);
        AmBBm = EngineHelper.buildAmBBm(ListUtils.subtract(Am, Bm), sources[0], principal);
        AAmBm = EngineHelper.buildAAmBm(ListUtils.subtract(Bm, Am), sources[1], principal);
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("AmBm:  " + AmBm );
            log.finest("AmBBm: " + AmBBm);
            log.finest("AAmBm: " + AAmBm);
        }
        
        //
        // Ready for conflict detection!
        //
        syncOperations = 
            checkSyncOperations(principal, lastSync, Am, Bm, AmBm, AmBBm, AAmBm);
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("operations: " + syncOperations);
        }
        
        if (log.isLoggable(Level.INFO)) {
            log.info("Preparation completed.");
        }
        
        return (SyncOperation[])syncOperations.toArray(new SyncOperationImpl[] {});
    }

    /**
     * Implements Synchronizable.sync
     *
     * 
     */
    public SyncOperationStatus[] sync(SyncOperation[] syncOperations) {
        Logger log = Logger.getLogger(LOG_NAME);
        
        if (log.isLoggable(Level.INFO)) {
            log.info("Synchronizing...");
        }
        
        if ((syncOperations == null) || (syncOperations.length == 0)) {
            return new SyncOperationStatus[0];
        }
        
        ArrayList status = new ArrayList();
        
        SyncOperationStatus[] operationStatus = null;
        for(int i=0; i<syncOperations.length; ++i) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Executing " + syncOperations[i]);
            }
            
            //
            // execSyncOperation can return more than one status for one 
            // operation when more than one source are involved
            //
            operationStatus = execSyncOperation((SyncOperationImpl)syncOperations[i]);
            
            for (int j=0; j<operationStatus.length; ++j) {
                status.add(operationStatus[j]);
            }  // next j
        } // next i
        
        if (log.isLoggable(Level.FINEST)) {
            log.finest("status: " + status);
        }

        return (SyncOperationStatus[])status.toArray(new SyncOperationStatus[0]);
    }
        
    /**
     * Implements Synchronizable.endSync
     *
     * @see beanblue.sync.event.Synchronizable
     */
    public void endSync() throws SyncException {
        Logger log = Logger.getLogger(LOG_NAME);
        
        if (log.isLoggable(Level.INFO)) {
            log.info("Synchronization completed.");
        }
    }
    
    // -------------------------------------------------------- Protected methds

    /**
     * Checks the given SyncItem lists and creates the needed SyncOperations
     * following the rules described in the class description and in the 
     * architecture document.
     *
     * @param principal who has requested the synchronization
     * @param lastSync timestamp of the last synchronization
     * @param Am the Am set
     * @param Bm the Bm set
     * @param AmBm the AmBm 
     * @param AmBBm the AmBBm
     * @param AAmBm the AAmBm
     *
     * @return an ArrayList containing all the collected sync operations
     */
    protected ArrayList checkSyncOperations(Principal principal,
                                            Timestamp lastSync ,
                                            List      Am       ,
                                            List      Bm       ,
                                            List      AmBm     ,
                                            List      AmBBm    , 
                                            List      AAmBm    ) {
        SyncItemMapping mapping = null;
        SyncItem syncItemA = null, syncItemB = null;
        
        ArrayList all        = new ArrayList();
        ArrayList operations = new ArrayList();
                
        // ---------------------------------------------------------------------
        
        all.addAll(AmBm );
        all.addAll(AmBBm);
        all.addAll(AAmBm);
        
        //
        // 1st check: items in both sources
        //
        Iterator i = all.iterator();
        while (i.hasNext()) {
            mapping = (SyncItemMapping)i.next();
            
            syncItemA = mapping.getSyncItemA();
            syncItemB = mapping.getSyncItemB();
            
            operations.add(
                checkSyncOperation(
                    principal,
                    lastSync , 
                    syncItemA,
                    syncItemB
                )
            );
            Am.remove(syncItemA);
            Bm.remove(syncItemB);
        }
        
        //
        // 2nd check: items in source A and not in source B
        //
        i = Am.iterator();
        while (i.hasNext()) {
           syncItemA = (SyncItem)i.next();
            
           operations.add(checkSyncOperation(principal, lastSync, syncItemA, null));
        }  // next i 
        
        //
        // 3rd check: items in source B and not in source A
        //
        i = Bm.iterator();
        while (i.hasNext()) {
           syncItemB = (SyncItem)i.next();
            
           operations.add(checkSyncOperation(principal, lastSync, null, syncItemB));
        }  // next i 
        
        return operations;
    }
    
    /**
     * Create a SyncOperation based on the state of the given SyncItem couple.
     *
     * @param principal     the entity that wnats to do the operation
     * @param since         timestamp of the last sync (user for "B" operation)
     * @param syncItemA     the SyncItem of the source A - NULL means <i>not existing</i/
     * @param syncItemB     the SyncItem of the source B - NULL means <i>not existing</i/
     *
     * @return the SyncOperation object
     */
    protected SyncOperation checkSyncOperation(Principal principal,
                                               Timestamp since    ,
                                               SyncItem  syncItemA, 
                                               SyncItem  syncItemB) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest( "check: syncItemA: "
                    + syncItemA
                    + " syncItemB: "
                    + syncItemB
                    );
        }
        
        if (syncItemA == null) {
            syncItemA = SyncItemImpl.getNotExistingSyncItem(null);
        }
        if (syncItemB == null) {
            syncItemB = SyncItemImpl.getNotExistingSyncItem(null);
        }
        
        switch (syncItemA.getState()) {
            //
            // NEW
            //
            case SyncItemState.NEW:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW) +
                                                String.valueOf(SyncItemState.NEW) );
                    case SyncItemState.UPDATED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW)     +
                                                String.valueOf(SyncItemState.UPDATED) );
                   case SyncItemState.DELETED:
                       return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW)     +
                                                String.valueOf(SyncItemState.DELETED) );
                   case SyncItemState.SYNCHRONIZED:
                       return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.NEW)       +
                                                String.valueOf(SyncItemState.SYNCHRONIZED) );
                   case SyncItemState.NOT_EXISTING:
                       syncItemA.setPropertyValue(syncItemB.PROPERTY_TIMESTAMP, since);
                       return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, false, true);
                }  // end inner switch
            
            //
            // DELETED
            //
            case SyncItemState.DELETED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.DELETED) +
                                                String.valueOf(SyncItemState.NEW)     );
                    case SyncItemState.UPDATED:
                        syncItemB.setState(SyncItemState.NEW);
                        return new SyncConflict(syncItemA, syncItemB, 
                                                String.valueOf(SyncItemState.DELETED) +
                                                String.valueOf(SyncItemState.NEW)     );
                   case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.DELETE, false, true);
                        
                   case SyncItemState.DELETED:
                   case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemB, syncItemA, SyncOperation.NOP, false, false);
                }  // end inner switch
            
            //
            // UPDATED
            //
            case SyncItemState.UPDATED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.NEW)     );
                    case SyncItemState.UPDATED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.UPDATED) );
                    case SyncItemState.DELETED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.DELETED) );
                    case SyncItemState.SYNCHRONIZED:
                        syncItemA.setPropertyValue(syncItemB.PROPERTY_TIMESTAMP, since);
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.UPDATE, false, true);
                    case SyncItemState.NOT_EXISTING:
                        syncItemA.setPropertyValue(syncItemB.PROPERTY_TIMESTAMP, since);
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, false, true);
                }  // end inner switch
            
           //
           // SYNCHRONIZED
           //
           case SyncItemState.SYNCHRONIZED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.SYNCHRONIZED) +
                                                String.valueOf(SyncItemState.NEW)       );
                    case SyncItemState.UPDATED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.UPDATE, true, false);
                    case SyncItemState.DELETED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.DELETE, true, false);
                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NOP, false, false);
                    case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, false, true);
                }  // end inner switch
                
           //
           // NOTEXISITNG
           //
           case SyncItemState.NOT_EXISTING:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                    case SyncItemState.UPDATED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, true, false);
                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NEW, true, false);
                    case SyncItemState.NOT_EXISTING:
                    case SyncItemState.DELETED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NOP, false, false);
                }  // end inner switch

           //
           // PARTIAL
           //
           case SyncItemState.PARTIAL:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.NEW)     );
                    case SyncItemState.UPDATED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.UPDATED) );
                    case SyncItemState.DELETED:
                        return new SyncConflict(syncItemA, syncItemB,
                                                String.valueOf(SyncItemState.UPDATED) +
                                                String.valueOf(SyncItemState.DELETED) );
                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, null, SyncOperation.ACCEPT_CHUNK, false, true);
                    case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemA, null, SyncOperation.ACCEPT_CHUNK, false, true);
                }  // end inner switch
                
            //
            // CONFLICT
            // In this case itemA has a mapped twin and itemB not existing
            //
            case SyncItemState.CONFLICT:
                return new SyncConflict(syncItemA, syncItemB, "CX");
        }  // end switch
        
        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NOP, false, false);
    }
    
    /**
     * Executes the given SyncOperation. Note that conflicts are ignored!
     * <p>
     * Note also that the number of status returned is equal to the number of
     * sources affected by the operation (0 for NOP, 1 for NEW and UPDATE and 
     * 1 or 2 for DELETE).
     *
     * @param operation     the SyncOperation to execute
     *
     * @return an array of <i>SyncOperationStatus</i> objects representing the 
     *         status of the executed operation. For instance, in case of error,
     *         <i>SyncOperation.error</i> will be set to the catched exception.
     */
    protected SyncOperationStatus[] execSyncOperation(SyncOperationImpl operation) {
        SyncItem syncItemA = operation.getSyncItemA(), 
                 syncItemB = operation.getSyncItemB();
        
        Principal owner = operation.getOwner();
        SyncOperationStatus[] status = null;
        ModificationCommand cmd = null;
        
        int size = 0, s = 0;

        switch (operation.getOperation()) {
            case SyncOperation.NEW:
                status = new SyncOperationStatus[1];
                if (operation.isAOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    try {
                        syncItemA = sources[1].setSyncItem(owner, syncItemB);
                        operation.setSyncItemA(syncItemA);
                        status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd, StatusCode.ITEM_ADDED);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                        operation.setAOperation(false); 
                        operation.setBOperation(false);
                    }
                } else if (operation.isBOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    syncItemB.setProperty(syncItemA.getProperty(SyncItem.PROPERTY_TIMESTAMP)); // this contains the
                                                                                               // current sync
                    try {
                        checkSize(syncItemA);
                        syncItemB = sources[0].setSyncItem(owner, syncItemA);
                        operation.setSyncItemB(syncItemB);                        
                        status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd, StatusCode.ITEM_ADDED);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[0] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                        operation.setAOperation(false); 
                        operation.setBOperation(false);
                    }
                }
                break;

            case SyncOperation.UPDATE:
                status = new SyncOperationStatus[1];
                if (operation.isAOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    try {
                        syncItemA = sources[1].setSyncItem(owner, syncItemB);
                        operation.setSyncItemA(syncItemA);
                        status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                        operation.setAOperation(false); 
                        operation.setBOperation(false);
                    }
                } else if (operation.isBOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    syncItemB.setProperty(syncItemA.getProperty(SyncItem.PROPERTY_TIMESTAMP)); // this contains the
                                                                                               // current sync
                    try {
                        checkSize(syncItemA);
                        syncItemB = sources[0].setSyncItem(owner, syncItemA);
                        operation.setSyncItemB(syncItemB);
                        status[0] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[0] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                        operation.setAOperation(false); 
                        operation.setBOperation(false);
                    }
                }
                break;

            case SyncOperation.DELETE:
                //
                // How many status we need? One if we have to delete the item
                // from only one source, two if we have to delete it from both
                // sources
                //
                size = 1;
                if (operation.isAOperation() && operation.isBOperation()) {
                    size = 2;
                }
                status = new SyncOperationStatus[size];
                                
                s = 0;
                if (operation.isBOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    syncItemB.setProperty(syncItemA.getProperty(SyncItem.PROPERTY_TIMESTAMP)); // this contains the
                                                                                               // timestamp of the 
                                                                                               // current sync
                    try {
                        sources[0].removeSyncItem(owner, syncItemB);
                        status[s++] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[s++] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                        operation.setAOperation(false); 
                        operation.setBOperation(false);
                    }
                } 
                if (operation.isAOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                                       
                    try {
                        sources[1].removeSyncItem(owner, syncItemA);
                        status[s++] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                    } catch (SyncException e) {
                        log.severe("Error executing sync operation: " + e.getMessage());
                        log.throwing(getClass().getName(), "execSyncOperation", e);
                        status[s++] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                        operation.setAOperation(false); 
                        operation.setBOperation(false);
                    }
                }
                break;

            case SyncOperation.NOP:
                //
                // A NOP operation can be due by one of the following conditions
                // (see checkOperation(...)):
                //
                // 1. both items are flagged as "Deleted"
                // 2. both items are flagged "Synchronized"
                // 3. itemA is "Not existing" and itemB is "Not existing" or "Deleted"
                //
                // In case 1. we want a status is returned, thought no real 
                // action is performed.
                //
                s = 0;
                if (operation.isAOperation()) ++size;
                if (operation.isBOperation()) ++size;
                
                status = new SyncOperationStatus[size];
                
                if (operation.isBOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    status[s++] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
                }
                
                if (operation.isAOperation()) {
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                    status[s++] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                }
                
                if (!operation.isAOperation() && !operation.isBOperation()) {
                    char stateItemA = syncItemA.getState();
                    char stateItemB = syncItemB.getState();
                    
                    switch (stateItemA) {
                        case SyncItemState.SYNCHRONIZED:
                            if (stateItemB == SyncItemState.SYNCHRONIZED) {
                                status = new SyncOperationStatus[1];
                                cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                                
                                if (cmd instanceof Add) {
                                    status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd, StatusCode.ITEM_ADDED);
                                } else {
                                    status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                                }
                                break;
                            }
                        case SyncItemState.NOT_EXISTING:
                            if (stateItemB == SyncItemState.DELETED) {
                                //
                                //client sent Delete item that not exist on server
                                //
                                try {
                                    cmd = (ModificationCommand)syncItemB.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                                    syncItemA = sources[1].setSyncItem(owner, syncItemB);
                                    operation.setSyncItemA(syncItemA);

                                    status = new SyncOperationStatus[1];
                                    status[0] = new Sync4jOperationStatusOK(
                                                operation, 
                                                sources[1],
                                                cmd, 
                                                StatusCode.ITEM_NOT_DELETED
                                                );
                                    break;
                                } catch (SyncException e) {
                                    log.severe("Error executing sync operation: " + e.getMessage());
                                    log.throwing(getClass().getName(), "execSyncOperation", e);
                                    status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                                    operation.setAOperation(false); 
                                    operation.setBOperation(false);
                                }
                            }
                        case SyncItemState.DELETED:
                            if (stateItemB == SyncItemState.DELETED) {
                                status = new SyncOperationStatus[1];
                                cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                                status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                                break;
                            }
                    }
                }
                                
                break;
                
            case SyncOperation.ACCEPT_CHUNK:
                status = new SyncOperationStatus[1];
                cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);
                status[0] = new Sync4jOperationStatusOK(
                                operation, sources[0], cmd, StatusCode.CHUNKED_ITEM_ACCEPTED
                            );
                break;

            case SyncOperation.CONFLICT:
                //
                // The conflic is solved with data from server
                //
                try {
                    SyncConflict sc = (SyncConflict)operation;                    
                    cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemHelper.PROPERTY_COMMAND);

                    //
                    //Type CX means that itemA is in state CONFLICT and itemB is in state NOT_EXISTING
                    //This situation happens when itemA has already a twin mapped
                    //
                    SyncItem resolvingItem = null;
                    if (sc.getType().equals(sc.STATE_CONFLICT_NONE)) {
                        resolvingItem = new SyncItemImpl(syncItemA.getSyncSource(), syncItemA.getMappedKey());
                        resolvingItem.setProperties(syncItemA.getProperties());
                        sources[1].removeSyncItem(owner, resolvingItem);
                    } else if (sc.getType().equals(sc.STATE_NEW_NEW)) {
                        //
                        // In this case, the items won't be mapped, therefore
                        // because server wins, the server item is returned
                        //
                        resolvingItem = new SyncItemImpl(
                                            syncItemB.getSyncSource(),
                                            syncItemB.getKey().getKeyAsString(),
                                            syncItemB.getKey().getKeyAsString(),
                                            SyncItemState.UNKNOWN
                        );
                        resolvingItem.setProperties(syncItemB.getProperties());
                        
                        syncItemA = sources[1].setSyncItem(owner, resolvingItem);
                        operation.setSyncItemA(syncItemA);
                    } else {
                        SyncItemKey luid = syncItemA.getMappedKey();
                        if (luid == null) {
                            throw new SyncException( "Item A with key " 
                                                   + syncItemA.getKey()
                                                   + " should having a mapping, instead the mapped key is null"
                                                   );
                        }
                        resolvingItem = new SyncItemImpl(
                                            syncItemB.getSyncSource(),
                                            syncItemA.getMappedKey().getKeyAsString()
                        );
                        resolvingItem.setProperties(syncItemB.getProperties());
                        
                        syncItemA = sources[1].setSyncItem(owner, resolvingItem);
                        operation.setSyncItemA(syncItemA);
                    }
                    operation.setAOperation(true);
                    status = new SyncOperationStatus[1];
                    
                    if (syncItemA.getState() == SyncItemState.NEW         &&
                        syncItemB.getState() == SyncItemState.SYNCHRONIZED
                       ) {                        
                        //
                        // This happens when the Client sent an item that have a twin
                        // synchronized on server.
                        //                        
                        status[0] = new Sync4jOperationStatusConflict(
                                    operation, 
                                    sources[1],
                                    cmd, 
                                    StatusCode.ALREADY_EXISTS
                                    );
                        break;
                    }
                    //
                    // in all the others cases...
                    //
                    status[0] = new Sync4jOperationStatusConflict(
                                    operation, 
                                    sources[1], 
                                    cmd, 
                                    StatusCode.CONFLICT_RESOLVED_WITH_SERVER_DATA
                                );
                } catch (SyncException e) {
                    log.severe("Error executing sync operation: " + e.getMessage());
                    log.throwing(getClass().getName(), "execSyncOperation", e);
                    status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                    operation.setAOperation(false); 
                    operation.setBOperation(false);
                }
                
                break;
        }  // end switch
        
        return status;
    }
    
    // --------------------------------------------------------- Private methods
    
    /**
     * If any item from the client has not a corresponding mapping, the 
     * server source must be queried for the item, since the client item 
     * could be the same of an existing server item. In this case, the old
     * unmapped item is replaced in Am by the newly mapped item.
     *
     * @param newlyMappedItems the collection that will contain the newly mapped
     *        items. If null, it won't be used.
     * @param syncItems the items to be searched if not mapped
     * @param source the source that has to be queried for twins
     * @param principal the principal items are belonging to
     *
     */
    private void fixMappedItems(Collection newlyMappedItems,
                                SyncItem[] syncItems       , 
                                SyncSource source          ,
                                Principal  principal       ) {        
        SyncItem itemB = null;
        
        for (int i = 0; ((syncItems != null) && (i<syncItems.length)); ++i) {
            itemB = null;
            if (!syncItems[i].isMapped()) {
                try {
                    itemB = source.getSyncItemFromTwin(principal, syncItems[i]);
                } catch (SyncSourceException e) {
                      String msg = "Error retrieving the twin item of "
                                 + syncItems[i].getKey()
                                 + " from source "
                                 + source
                                 + ": "
                                 + e.getMessage();
                      log.severe(msg);
                      log.throwing(getClass().getName(), "fixSyncMapping", e);
                }
                if (itemB != null) {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest( "Hey, client item "
                                  + syncItems[i].getKey().getKeyAsString()
                                  + " is the same of server item "
                                  + itemB.getKey().getKeyAsString()
                                  + '!'
                                  );
                    }
                    SyncItemMapping mapping = new SyncItemMapping(itemB.getKey());
                    mapping.setMapping(syncItems[i], itemB);
                    if (newlyMappedItems != null) {
                        newlyMappedItems.add(mapping);
                    }
                    syncItems[i] = SyncItemHelper.newMappedSyncItem(itemB.getKey(), syncItems[i]);
                }
            }
        }  // next i
    }
    
    /**
     * Verify if input item is already present into list of mapped items
     * 
     * @param newlyMappedItems list updated of items
     * @param itemTwin item to research
     *
     * @return true if item presents into list else false 
     */
    private boolean isContained(Collection newlyMappedItems, SyncItem itemTwin) {
        Iterator it2 = newlyMappedItems.iterator();
        while (it2.hasNext()) {
            SyncItemMapping map = (SyncItemMapping)it2.next();
            SyncItem itemB = map.getSyncItemB();                            

            if (itemB.equals(itemTwin)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Verifies that the given item the size of its content metches the 
     * specified size, if specified. In the case of a mismatch, a 
     * <code>SyncSourceException</code> is thrown with status code 
     * OBJECT_SIZE_MISMATCH (424).
     *
     * @param item item to check
     *
     * @throws SyncSourceException if the content size does not match the declared size
     *         (if declared)
     */
    private void checkSize(SyncItem item)
    throws SyncSourceException {
        Long size = (Long)item.getPropertyValue(item.PROPERTY_SIZE);
        
        if (size == null) {
            return;
        }
        
        byte[] data = (byte[])item.getPropertyValue(item.PROPERTY_BINARY_CONTENT);
        
        if (data.length != size.intValue()) {
            throw new SyncSourceException("The size of the received object does not match the given size", StatusCode.OBJECT_SIZE_MISMATCH);
        }
    }
}

