/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.transport.http.server;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.logging.Logger;
import java.util.logging.Level;

import javax.servlet.*;
import javax.servlet.http.*;

import sync4j.framework.config.Configuration;
import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.core.*;
import sync4j.framework.tools.Base64;
import sync4j.framework.server.*;
import sync4j.framework.tools.WBXMLTools;
import sync4j.framework.transport.http.SyncHolder;
import sync4j.framework.protocol.ProtocolException;

import org.apache.commons.lang.StringUtils;

import sync4j.transport.http.server.Constants;

import org.jibx.runtime.*;
import org.jibx.runtime.impl.*;

/**
 *  Receives the HTTP request and does session management.
 *
 *  <b>Session Management</b>
 *  <p>
 *  If the url contains a parameter {PARAM_SESSION_ID}={SESSION_ID}, the session
 *  id is used to lookup in the <i>handlerCache</i> if there is already a
 *  <i>SyncHolder</i> associated with it; the found handler is then used to process
 *  the incoming request. Otherwise, (either if the session parameter is not
 *  specified or no cached handler is found), a new session id is created and a
 *  new <i>SyncHolder</i> object is instantiated and stored in the cache.<br>
 *  The session id is created as followed:
 *  <p>
 *  <ol>
 *    <li> the first four bytes contains the IP address of the remote client
 *    <li> a dash ('-') is appended
 *    <li> the creation timestamp is appended
 *    <li> the resulting string is encoded base 64
 *  </ol>
 *  Note that no session expiration is handled by this class. That is delegated
 *  to the <i>SyncHandler</i> object.
 *
 *  @author Stefano Fornari
 *
 *  @version $Id: ClusterableSync4jServlet.java,v 1.7.4.8 2005/01/26 16:55:06 stefano_fornari Exp $
 *
 */
public final class ClusterableSync4jServlet
extends HttpServlet
implements Constants {
    // ------------------------------------------------------------ Private data
    private String jndiAddress = DEFAULT_JNDI_ADDRESS;
    private RemoteHolderCache holderCache = null;

    private Logger log = Logger.getLogger(LOG_NAME);

    //
    // To check if the engine was initialized
    //
    private boolean initialized = false;
    private static final String PATH_STATUS = "/sync4j/sync/status";

    // ---------------------------------------------------------- Public methods

    public void init() throws ServletException {
		initialized = false;
        String value;

        ServletConfig config = getServletConfig();

        value = config.getInitParameter(PARAM_CHANNEL_PROPERTIES);
        if (StringUtils.isEmpty(value)) {
            String msg = "The servlet configuration parameter "
                       + PARAM_CHANNEL_PROPERTIES
                       + " cannot be empty ("
                       + value
                       + ")"
                       ;
            log.severe(msg);
            throw new ServletException(msg);
        }

        value = config.getInitParameter(PARAM_SESSION_TTL);
        if (!StringUtils.isEmpty(value)) {
            try {
                Long.parseLong(value);
            } catch (NumberFormatException e) {
                String msg = "The servlet configuration parameter "
                           + PARAM_SESSION_TTL
                           + " is not an integer number ("
                           + value
                           + ")"
                           ;
                log.severe(msg);
                throw new ServletException(msg);
            }
        }

        value = config.getInitParameter(PARAM_JNDI_ADDRESS);
        if (StringUtils.isEmpty(value)) {
            String msg = "Missing optional parameter "
                       + PARAM_JNDI_ADDRESS
                       + ", default value ("
                       + jndiAddress
                       + ") used."
                       ;
            log.warning(msg);
        } else {
            jndiAddress = value;
        }

        holderCache = new RemoteHolderCache(config);

        //
        // Now we are ready to bootstrap the Sync4j components
        //
        bootstrap();

        //
        // Engine initialized!
        //
        initialized = true;
    }


    public void doPost(
            final HttpServletRequest request,
            final HttpServletResponse response)
            throws ServletException, IOException {

        if (log.isLoggable(Level.INFO)) {
            StringBuffer sb = new StringBuffer("Handling incoming request ");

            sb.append(request.getContextPath())
              .append(request.getServletPath());
            if (request.getPathInfo() != null) {
                sb.append(request.getPathInfo());
            }

            if (request.getQueryString() != null) {
                sb.append('?').append(request.getQueryString());
            }
            sb.append('.');

            log.info(sb.toString());
        }

        final String contentType    = request.getContentType().split(";")[0];
        final int    contentLength  = request.getContentLength();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("contentType: " + contentType);
            log.finest("contentLength: " + contentLength);
        }

        if (contentLength < 1) {
            throw new Error("Content length < 1 (" + contentLength + ")");
        }

        byte[] requestData = new byte[contentLength];

        InputStream in = request.getInputStream();
        int n = 0;
        int bytesRead = 0;

        try {
            do {
                n = in.read(requestData,
                            bytesRead,
                            requestData.length - bytesRead);
                if (n > 0) {
                    bytesRead += n;
                }
            } while (n != -1);
        } catch (IOException ex) {
            handleError(request, response, "Error reading the request", ex);
            return;
        } finally {
            try {
                if (in != null) {
                    in.close();
                }
            } catch (Exception ex) {
                handleError(request, response, ex.getClass().getName(), ex);
                return;
            }
            in = null;
        }

        if (bytesRead != contentLength) {
            handleError(request, response, "bytesRead != contentLength", null);
            return;
        }

        //
        // If the session id is not specified in the URL, a new remote object
        // will be created. Otherwise the session id specifies which remote
        // object shall handles the request.
        //
        RemoteEJBSyncHolder holder = null;
        try {
            holder = createHolder(request);
        } catch (Exception e) {
            handleError(request, response, "Error creating SyncBean", e);
            return;
        }

        SyncResponse resp = null;
        try {
            resp = holder.processMessage(requestData, contentType);
        } catch (Exception e) {
            Throwable cause = e.getCause();

            if (  (cause != null)
               && (  (cause instanceof ProtocolException)
                  || (cause instanceof Sync4jException  )
                  )
               ) {
                handleError(request, response, "Protocol error", cause);
                return;
            } else {
                throw new ServletException(e);
            }
        } finally {
            //
            // Now we can store the handler into the distributed cache. Note
            // that when the holder is put in the cache a serialized copy is
            // actually stored, not the simple local reference.
            //
            if (holder.isNew()) {
                holderCache.put(holder);
            }
        }

        OutputStream out = null;
        try {
            out = response.getOutputStream();
            response.setContentType(contentType);

            byte[] buf = resp.getMessage();
            response.setContentLength(buf.length);
            out.write(buf);
            out.flush();

        } finally {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Finally");
            }
            if (out != null) {
                out.close();
            }
        }

        //
        // If the message completed the SyncML communication, the holder
        // must be closed and discarded.
        //
        if (resp.isCompleted()) {
            releaseHolder(holder);
        }
    }

    public void doGet(
            final HttpServletRequest request,
            final HttpServletResponse response)
            throws ServletException, IOException {
        if (log.isLoggable(Level.INFO)) {
        log.info("holderCache: " + holderCache);
        }
        super.doGet(request, response);

        String requestURI = request.getRequestURI();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("requestURI: " + requestURI);
        }
        if (requestURI.equals(PATH_STATUS)) {
            if (initialized) {
                //
                //The request sent by the client was successful
                //
                response.setStatus(response.SC_OK);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Request succeeded normally");
                }
            } else {
                //
                //The request was unsuccessful to the server being down or overloaded.
                //
                response.setStatus(response.SC_SERVICE_UNAVAILABLE);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("The HTTP server is unable to handle the request");
                }
            }
        }
    }

    // --------------------------------------------------------- Private methods


    /**
     * Factory method for <i>SyncHolder</i> objects. If the session id is
     * passed as a CGI parameter, it is transformed in a EJB handle. If the
     * handle is not valid, it is considered expired. If the session id is not
     * specified, a new EJB is created
     *
     * @param request the associated HTTP request object
     *
     * @return a new <i>SyncHolder</i>
     *
     */
    private RemoteEJBSyncHolder createHolder(HttpServletRequest request)
    throws Exception {
        String sessionId = request.getParameter(PARAM_SESSION_ID);

        RemoteEJBSyncHolder holder = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("holderCache: " + holderCache);
        }

        if (!StringUtils.isEmpty(sessionId)) {
            holder = (RemoteEJBSyncHolder)holderCache.get(sessionId);
        }

        if (holder == null) {
            holder = new RemoteEJBSyncHolder(true);

            holder.setSessionId(createSessionId(request));
            holder.setJndiAddress(jndiAddress);
        }

        return holder;
    }

    /**
     * Closes the given <i>SyncHolder</i> and removes it from the cache.
     *
     * @param holder the holder to releases
     */
    private void releaseHolder(SyncHolder holder) {
        try {
            holder.close();
        } catch (Exception e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "releaseHolder", e.getCause());
        }
        holderCache.remove(holder.getSessionId());
    }

    /**
     * Handles errors conditions returning an appropriate content to the client.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest   request,
                             final HttpServletResponse response,
                                   String                   msg,
                             final Throwable                  t) {

        if (msg == null) {
            msg = "";
        }

        if (t == null) {
           log.severe(msg);
        } else {
            log.severe(msg);
            log.throwing(getClass().getName(), "handleError", t);
        }
        try {
            response.sendError(response.SC_BAD_REQUEST, msg);
        } catch (IOException e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "handleError", e);
        }
    }

    /**
     * Creates the session id (see the class description for details).
     *
     * @param request the HTTP request object
     *
     * @return a newly created session id
     */
    private String createSessionId(HttpServletRequest request) {
        String clientIP  = request.getRemoteAddr()   ;
        long   timestamp = System.currentTimeMillis();

        StringBuffer sb = new StringBuffer();

        byte[] ip = null;

        try {
            ip = InetAddress.getByName(clientIP).getAddress();
        } catch (UnknownHostException e) {
            ip = new byte[] {0, 0, 0, 0};
        }

        sb.append(ip).append('-').append(timestamp);

        String stringSessionId = new String(Base64.encode(sb.toString().getBytes()));

        //
        // strips out tailing '='
        //
        int i = stringSessionId.lastIndexOf('=');

        return ((i>0) ? stringSessionId.substring(0, i-1) : stringSessionId);
    }

    /**
     * The same as <i>handleError(reqest, response, msg, t, null)</i>.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest   request,
                             final HttpServletResponse response,
                             final String                   msg) {
        this.handleError(request, response, msg, null);
    }

    /**
     * Bootstraps the Sync4j server components and logs the server startup
     * event on both the console and the logging system.
     */
    private void bootstrap() {
        Configuration c = null;

        c = Configuration.getConfiguration();

        String msg1 = "================================================================================";
        String sp   = "\n\n";
        String msg2 = "SyncServer v. "
                    + c.getStringValue(ConfigurationConstants.CFG_ENGINE_SWVERSION)
                    + " engine started.\nConfiguration object found."
                    ;

        //
        // NOTE: this code uses the stdout by choice, so that it will always
        // be displayed, regardless how logging is configured.
        //
        System.out.println(msg1);
        System.out.println(sp);
        System.out.println(msg2);
        System.out.println(sp);
        System.getProperties().list(System.out);
        System.out.println(sp);
        System.out.println(msg1);

        if (log.isLoggable(Level.INFO)) {
            log.info(msg1+sp);
            log.info(msg2+sp);
            log.info(System.getProperties().toString()+sp);
            log.info(msg1);
        }
    }

}